  SUBROUTINE Cell_face_velocity_FV_3D(Nx , Ny , NZ ,  C_X , C_Y , C_Z ,  Uc , Vc , Wc)
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !Routine to compute cell-face velocity on uniform cell-center FV.
   !Compute cell-face velocity using simple average
   !
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER,INTENT(IN) :: Nx , Ny  ,NZ
     REAL(KIND=DP),INTENT(IN ) :: C_X(:,:,:) , C_Y(:,:,:) , C_Z(:,:,:) 
     REAL(KIND=DP),INTENT(OUT) ::  Uc(:,:,:) ,  Vc(:,:,:) ,  Wc(:,:,:) 
 
     INTEGER :: I , J ,  K 
 
     Uc = 0.0_DP
     Vc = 0.0_DP
     Wc = 0.0_DP   
 
 
    
      !Compute cell face value Uc , Vc
     !For Uc
     DO K = 2 , NZ-1
     DO J = 2 , Ny-1
       Uc(1,J,K)    = C_X(1,J,K)
       Uc(Nx-1,J,K) = C_X(Nx,J,K)
     ENDDO
     ENDDO
    
      DO K = 2 , NZ-1   
      DO J = 2 , Ny-1   
      DO I = 2 , Nx-2
         Uc(I,J,K) = 0.5_DP * (C_X(I+1,J,K) + C_X(I,J,K))
      ENDDO
      ENDDO
      ENDDO
 
     !For Vc
     DO K = 2 , NZ-1 
     DO I = 2 , Nx-1
       Vc(I,1,K)    = C_Y(I,1,K)
       Vc(I,Ny-1,K) = C_Y(I,Ny,K)
     ENDDO
     ENDDO
  
     DO K = 2 , NZ-1
     DO J = 2 , Ny-2 
     DO I = 2 , Nx-1
         Vc(I,J,K) = 0.5_DP * (C_Y(I,J+1,K) + C_Y(I,J,K))
     ENDDO
     ENDDO
     ENDDO
     
     
      !For Wc
     DO J = 2 , NY-1 
     DO I = 2 , Nx-1
       Wc(I,J,1)    = C_Z(I,J,1)
       Wc(I,J,NZ-1) = C_Z(I,J,NZ)
     ENDDO
     ENDDO
  
     DO K = 2 , NZ-2 
     DO I = 2 , Nx-1
     DO J = 2 , Ny-1
         Wc(I,J,K) = 0.5_DP * (C_Z(I,J,K+1) + C_Z(I,J,K))
     ENDDO
     ENDDO
     ENDDO
     
     
 
   END SUBROUTINE Cell_face_velocity_FV_3D
   
  
  
   SUBROUTINE FIND_FxFy_001_FV_3D(N , M , L  , Fx, Fy, FZ , FF, Dx , Dy, DZ)
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !Routine to Get First derivatives with uniform Grids.
   !
   !
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     IMPLICIT NONE
 
     INTEGER      , INTENT(IN)  :: N , M , L
     REAL(KIND=DP), INTENT(IN)  :: Dx , Dy , DZ
     REAL(KIND=DP), INTENT(OUT) :: Fx(:,:,:) , Fy(:,:,:) , FZ(:,:,:)
     REAL(KIND=DP), INTENT(IN)  :: FF(:,:,:)
     
     REAL(KIND=DP) :: F(N,M,L)
     
     INTEGER       :: I, J, K
 
     
     DO K = 1 , L
     DO J = 1 , M
     DO I = 1 , N
       F(I,J,K) = FF(I,J,K)
     END DO
     END DO
     END DO
 
     !Extrapolate value to ghost cell.
     DO K = 1 , L
     DO J = 1 , M
       F(1,J,K)  = 2.0_DP * FF(1,J,K)  - FF(2,J,K)
       F(N,J,K) = 2.0_DP * FF(N,J,K) - FF(N-1,J,K)
     END DO
     END DO
 
 
     DO K = 1 , L
     DO I = 1 , N      
       F(I,1,K)  = 2.0_DP * FF(I,1,K)  - FF(I,2,K)
       F(I,M,K) = 2.0_DP * FF(I,M,K) - FF(I,M-1,K) 
     END DO
     END DO
     
     
     DO J = 1 , M
     DO I = 1 , N      
       F(I,J,1)  = 2.0_DP * FF(I,J,1)  - FF(I,J,2)
       F(I,J,L) = 2.0_DP * FF(I,J,L) - FF(I,J,L-1) 
     END DO
     END DO
 
     
    F(1,1,1)   = 2.0_DP * F(1,1,1)   -  F(2,2,2)       
    F(1,m,1)  = 2.0_DP * F(1,M,1)  -  F(2,M-1,2)   
    F(N,1,1)  = 2.0_DP * F(N,1,1)  -  F(N-1,2,2)    
    F(N,m,1) = 2.0_DP * F(N,M,1) -  F(N-1,M-1,2) 

    F(1,1,L)   = 2.0_DP * F(1,1,L)   - F(2,2,L-1)    
    F(1,M,L)  = 2.0_DP * F(1,M,L)  -  F(2,M-1,L-1)   
    F(N,1,L)  = 2.0_DP * F(N,1,L)  -  F(N-1,2,L-1)   
    F(N,M,L) = 2.0_DP * F(N,M,L) -  F(N-1,M-1,L-1) 
 
    
        
    DO K = 1 , L
    DO J = 1 , M
    DO I = 1 , N
      Fx(I,J,K) = 0.0_DP
      Fy(I,J,K) = 0.0_DP
      FZ(I,J,K) = 0.0_DP
    END DO
    END DO
    END DO
 

 
     DO K = 2 , L-1
     DO J = 2 , M-1
     DO I = 2 , N-1
       Fx(I,J,K) = (F(I+1,J,K) - F(I-1,J,K)) / (2.0_DP*Dx)
       Fy(I,J,K) = (F(I,J+1,K) - F(I,J-1,K)) / (2.0_DP*Dy)
       FZ(I,J,K) = (F(I,J,K+1) - F(I,J,K-1)) / (2.0_DP*DZ)
     END DO
     END DO
     END DO
 
 
   END SUBROUTINE FIND_FxFy_001_FV_3D
   
   
   SUBROUTINE FIND_FxFy_002_FV_3D(N , M , L ,  Fxx , Fyy  , FZZ , FF, DX, DY  , DZ)
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !Routine to Get Second derivatives with uniform Grids.
   !
   !
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     IMPLICIT NONE
 
     INTEGER      , INTENT(IN)  :: N , M , L
     REAL(KIND=DP), INTENT(IN)  :: DX , DY , DZ
     REAL(KIND=DP), INTENT(OUT) :: Fxx(:,:,:) , Fyy(:,:,:) , FZZ(:,:,:)
     REAL(KIND=DP), INTENT(IN)  :: FF(:,:,:)
     REAL(KIND=DP) :: F(N,M,L)
     INTEGER                    :: I , J , K
      
     DO K = 1 , L 
     DO J = 1 , M
     DO I = 1 , N
       Fxx(I,J,K) = 0.0_DP
       Fyy(I,J,K) = 0.0_DP
       FZZ(I,J,K) = 0.0_DP
     END DO
     END DO
     END DO
     
     DO K = 1 , L
     DO J = 1 , M
     DO I = 1 , N
       F(I,J,K) = FF(I,J,K)
     END DO
     END DO
     END DO
 
     !Extrapolate value to ghost cell.
     DO K = 1 , L
     DO J = 1 , M
       F(1,J,K)  = 2.0_DP * FF(1,J,K)  - FF(2,J,K)
       F(N,J,K) = 2.0_DP * FF(N,J,K) - FF(N-1,J,K)
     END DO
     END DO
 
 
     DO K = 1 , L
     DO I = 1 , N      
       F(I,1,K)  = 2.0_DP * FF(I,1,K)  - FF(I,2,K)
       F(I,M,K) = 2.0_DP * FF(I,M,K) - FF(I,M-1,K) 
     END DO
     END DO
     
     
     DO J = 1 , M
     DO I = 1 , N      
       F(I,J,1)  = 2.0_DP * FF(I,J,1)  - FF(I,J,2)
       F(I,J,L) = 2.0_DP * FF(I,J,L) - FF(I,J,L-1) 
     END DO
     END DO
 
     
    F(1,1,1)   = 2.0_DP * F(1,1,1)   -  F(2,2,2)       
    F(1,m,1)  = 2.0_DP * F(1,M,1)  -  F(2,M-1,2)   
    F(N,1,1)  = 2.0_DP * F(N,1,1)  -  F(N-1,2,2)    
    F(N,m,1) = 2.0_DP * F(N,M,1) -  F(N-1,M-1,2) 

    F(1,1,L)   = 2.0_DP * F(1,1,L)   - F(2,2,L-1)    
    F(1,M,L)  = 2.0_DP * F(1,M,L)  -  F(2,M-1,L-1)   
    F(N,1,L)  = 2.0_DP * F(N,1,L)  -  F(N-1,2,L-1)   
    F(N,M,L) = 2.0_DP * F(N,M,L) -  F(N-1,M-1,L-1) 
    
    
    DO K = 2 , L-1
    DO J = 2 , M-1
    DO I = 2 , N-1
       Fxx(I,J,K) = (F(I+1,J,K) -2.0_DP*F(I,J,K) + F(I-1,J,K)) / (Dx**2)
       Fyy(I,J,K) = (F(I,J+1,K) -2.0_DP*F(I,J,K) + F(I,J-1,K)) / (Dy**2)
       FZZ(I,J,K) = (F(I,J,K+1) -2.0_DP*F(I,J,K) + F(I,J,K-1)) / (DZ**2)
    END DO
    END DO
    END DO
 
   END SUBROUTINE FIND_FxFy_002_FV_3D
   
   
  SUBROUTINE FIND_Fxy_2nd_fv_3D(N , M  , L , Fxy , FYZ  , FXZ  , FF , Dx , Dy , DZ)
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !Routine to Get Mixed derivatives with uniform Grids.
   !
   !F_XY , F_YZ , F_XZ
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER,PARAMETER :: DP = SELECTED_REAL_KIND(p=15)
 
     INTEGER      , INTENT(IN)  :: N , M  , L
     REAL(KIND=DP), INTENT(IN)  :: Dx , Dy  , DZ 
     REAL(KIND=DP), INTENT(OUT) :: Fxy(:,:,:) ,  FYZ(:,:,:) ,  FxZ(:,:,:)
     REAL(KIND=DP), INTENT(IN)  :: FF(:,:,:)
     REAL(KIND=DP)              :: F(N,M,L)
     INTEGER                    :: I , J , K
 
  
     Fxy = 0.0_DP
     FyZ = 0.0_DP  
     FxZ = 0.0_DP  
    
     DO K = 1 , L
     DO J = 1 , M
     DO I = 1 , N
       F(I,J,K) = FF(I,J,K)
     END DO
     END DO
     END DO
 
     !Extrapolate value to ghost cell.
     DO K = 1 , L
     DO J = 1 , M
       F(1,J,K)  = 2.0_DP * FF(1,J,K)  - FF(2,J,K)
       F(N,J,K) = 2.0_DP * FF(N,J,K) - FF(N-1,J,K)
     END DO
     END DO
 
 
     DO K = 1 , L
     DO I = 1 , N      
       F(I,1,K)  = 2.0_DP * FF(I,1,K)  - FF(I,2,K)
       F(I,M,K) = 2.0_DP * FF(I,M,K) - FF(I,M-1,K) 
     END DO
     END DO
     
     
     DO J = 1 , M
     DO I = 1 , N      
       F(I,J,1)  = 2.0_DP * FF(I,J,1)  - FF(I,J,2)
       F(I,J,L) = 2.0_DP * FF(I,J,L) - FF(I,J,L-1) 
     END DO
     END DO
 
     
    F(1,1,1)   = 2.0_DP * F(1,1,1)   -  F(2,2,2)       
    F(1,m,1)  = 2.0_DP * F(1,M,1)  -  F(2,M-1,2)   
    F(N,1,1)  = 2.0_DP * F(N,1,1)  -  F(N-1,2,2)    
    F(N,m,1) = 2.0_DP * F(N,M,1) -  F(N-1,M-1,2) 

    F(1,1,L)   = 2.0_DP * F(1,1,L)   - F(2,2,L-1)    
    F(1,M,L)  = 2.0_DP * F(1,M,L)  -  F(2,M-1,L-1)   
    F(N,1,L)  = 2.0_DP * F(N,1,L)  -  F(N-1,2,L-1)   
    F(N,M,L) = 2.0_DP * F(N,M,L) -  F(N-1,M-1,L-1) 
 
 
 
     !inner
     DO K = 2 , L-1
     DO J = 2 , M-1
     DO I = 2 , N-1
        Fxy(I,J,K) = ( (F(I+1,J+1,K) - F(I+1,J-1,K)) / (2.0_DP*Dy) &
                &  -(F(I-1,J+1,K) - F(I-1,J-1,K)) / (2.0_DP*Dy) ) / (2.0_DP * Dx)
     END DO
     END DO
     END DO
     
     DO K = 2 , L-1
     DO J = 2 , M-1
     DO I = 2 , N-1
        FxZ(I,J,K) = ( (F(I+1,J,K+1) - F(I+1,J,K-1)) / (2.0_DP*DZ) &
                &  -(F(I-1,J,K+1) - F(I-1,J,K-1)) / (2.0_DP*DZ) ) / (2.0_DP * Dx)
     END DO
     END DO
     END DO
     
     
     DO K = 2 , L-1
     DO J = 2 , M-1
     DO I = 2 , N-1
        FYZ(I,J,K) = ( (F(I,J+1,K+1) - F(I,J+1,K-1)) / (2.0_DP*DZ) &
                &  -(F(I,J-1,K+1) - F(I,J-1,K-1)) / (2.0_DP*DZ) ) / (2.0_DP * DY)
     END DO
     END DO
     END DO
 
    
 
   END SUBROUTINE FIND_Fxy_2nd_fv_3D
   
   
    SUBROUTINE Cell_face_derivative_FV_3D(Nx , Ny ,  NZ , Dx , Dy ,  DZ , U , U_X , U_Y , U_Z)
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    !Routine to compute cell face derivative on uniform cell-center FV.
    !compute using simple central difference idea.
    !
    !
    !
    !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
      INTEGER,INTENT(IN) :: Nx , Ny , NZ
   
      REAL(KIND=DP),INTENT(IN ) :: Dx , Dy , DZ , U(:,:,:)
      REAL(KIND=DP),INTENT(OUT) :: U_X(:,:,:) , U_Y(:,:,:) , U_Z(:,:,:)
   
      REAL(KIND=DP),DIMENSION(Nx,Ny,NZ) :: U_old
      INTEGER :: I , J  , K
   
   
      U_old(1:Nx,1:Ny,1:NZ) = U(1:Nx,1:Ny,1:NZ)
   
       DO K = 1 , NZ
       DO J = 1 , Ny
         U_old(1,J,K)  = ( 8.0_DP * U(1,J,K)  - 6.0_DP * U(2,J,K)    + U(3,J,K)    ) / 3.0_DP
         U_old(Nx,J,K) = ( 8.0_DP * U(Nx,J,K) - 6.0_DP * U(Nx-1,J,K) + U(Nx-2,J,K) ) / 3.0_DP      
    
       END DO
       END DO
         
       DO K = 1  ,NZ
       DO I = 1 , Nx
         U_old(I,1,K)  = ( 8.0_DP * U(I,1,K)  - 6.0_DP * U(I,2,K)    + U(I,3,K)    ) / 3.0_DP
         U_old(I,Ny,K) = ( 8.0_DP * U(I,Ny,K) - 6.0_DP * U(I,Ny-1,K) + U(I,Ny-2,K) ) / 3.0_DP
           
       END DO
       END DO
       
          
       DO J = 1 , NY
       DO I = 1 , Nx
         U_old(I,J,1)  = ( 8.0_DP * U(I,J,1)  - 6.0_DP * U(I,J,2)    + U(I,J,3)    ) / 3.0_DP
         U_old(I,J,NZ) = ( 8.0_DP * U(I,J,NZ) - 6.0_DP * U(I,J,NZ-1) + U(I,J,NZ-2) ) / 3.0_DP 
         
     
       END DO
       END DO          
       
       U_old(1,1,1)   = (8.0_DP * U(1,1,1)   - 6.0_DP * U(2,2,2)       + U(3,3,3)      ) / 3.0_DP
       U_old(1,Ny,1)  = (8.0_DP * U(1,Ny,1)  - 6.0_DP * U(2,Ny-1,2)    + U(3,Ny-2,3)   ) / 3.0_DP
       U_old(Nx,1,1)  = (8.0_DP * U(Nx,1,1)  - 6.0_DP * U(Nx-1,2,2)    + U(Nx-2,3,3)   ) / 3.0_DP
       U_old(Nx,Ny,1) = (8.0_DP * U(Nx,Ny,1) - 6.0_DP * U(Nx-1,Ny-1,2) + U(Nx-2,Ny-2,3)) / 3.0_DP
    
       U_old(1,1,NZ)   = (8.0_DP * U(1,1,NZ)   - 6.0_DP * U(2,2,NZ-1)       + U(3,3,NZ-2)      ) / 3.0_DP
       U_old(1,Ny,NZ)  = (8.0_DP * U(1,Ny,NZ)  - 6.0_DP * U(2,Ny-1,NZ-1)    + U(3,Ny-2,NZ-2)   ) / 3.0_DP
       U_old(Nx,1,NZ)  = (8.0_DP * U(Nx,1,NZ)  - 6.0_DP * U(Nx-1,2,NZ-1)    + U(Nx-2,3,NZ-2)   ) / 3.0_DP
       U_old(Nx,Ny,NZ) = (8.0_DP * U(Nx,Ny,NZ) - 6.0_DP * U(Nx-1,Ny-1,NZ-1) + U(Nx-2,Ny-2,NZ-2)) / 3.0_DP
   
   
      !Compute cell face derivative U_X , U_Y
      !For U_X
      
      DO K = 2 , NZ-1
      DO J = 2 , Ny-1          
      DO I = 1 , Nx-1
          U_X(I,J,K) = (U_old(I+1,J,K) - U_old(I,J,K)) / Dx
      END DO
      END DO
      END DO
   
      !For U_Y
      DO K = 2 , NZ-1
      DO J = 1 , Ny-1
      DO I = 2 , Nx-1  
          U_Y(I,J,K) = (U_old(I,J+1,K) - U_old(I,J,K)) / Dy
      END DO
      END DO
      END DO
      
      
      DO K = 1 , NZ-1
      DO J = 2 , Ny-1
      DO I = 2 , Nx-1  
          U_Z(I,J,K) = (U_old(I,J,K+1) - U_old(I,J,K)) / DZ
      END DO
      END DO
      END DO
   
    END SUBROUTINE Cell_face_derivative_FV_3D
   
   
  SUBROUTINE LS_SURFACE_FORCE_3D(N , M , L ,  DX , DY  , DZ , FHI , Kappa)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !
  !
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  INTEGER,PARAMETER  :: DP = SELECTED_REAL_KIND(P=15)
  
  INTEGER,INTENT(IN) :: N , M   , L
  
  REAL(KIND=DP),INTENT(IN)  :: FHI(:,:,:)
  REAL(KIND=DP),INTENT(OUT) :: Kappa(:,:,:)
  
  REAL(KIND=DP) :: F_X(N,M,L) , F_Y(N,M,L)  , F_Z(N,M,L), F_XX(N,M,L) , F_YY(N,M,L) , F_ZZ(N,M,L)
  REAL(KIND=DP) ::  F_XY(N,M,L) ,  F_YZ(N,M,L) , F_XZ(N,M,L)
  REAL(KIND=DP), INTENT(IN)  :: DX , DY , DZ
  
  Kappa = 0.0_DP

  CALL FIND_FxFy_001_fv_3D(N , M , L , F_X , F_Y  , F_Z , FHI, Dx, Dy , DZ)
  CALL FIND_FxFy_002_fv_3D(N , M  , L , F_XX , F_YY , F_ZZ , FHI, Dx, Dy , DZ) 
  CALL FIND_Fxy_2nd_fv_3D(N , M , L , F_XY , F_YZ , F_XZ , FHI , Dx, Dy , Dz)

  
  !Kappa = ( (F_X**2) * F_YY - 2.0_DP * F_X*F_Y*F_XY + (F_Y**2) * F_XX ) &
  !      & / (F_X**2 + F_Y**2)**1.5_DP
        
    
  Kappa = ( (F_X**2) * F_YY - 2.0_DP * F_X*F_Y*F_XY + (F_Y**2) * F_XX +   &
              (F_X**2) * F_ZZ - 2.0_DP * F_X*F_Z*F_XZ + (F_Z**2) * F_XX +   &
              (F_Y**2) * F_ZZ - 2.0_DP * F_Y*F_Z*F_YZ + (F_Z**2) * F_YY   )&
          & / (F_X**2 + F_Y**2 + F_Z**2)**1.5_DP  
    
        
  ! DO J = 2 , M-1
  !    KAPPA(1,J)  = KAPPA(2,J)
  !    KAPPA(N,J) =  KAPPA(N-1,J)
  ! END DO
  !
  !  DO I = 2 , N-1
  !    KAPPA(I,1)  =  KAPPA(I,2)
  !    KAPPA(I,M) =  KAPPA(I,M-1)
  ! END DO
  !
  !  KAPPA(1,1)  =   KAPPA(2,2)
  !  KAPPA(1,M)  =   KAPPA(2,M-1)
  !  KAPPA(N,1)  =   KAPPA(N-1,2)
  !  KAPPA(N,M)  =   KAPPA(N-1,M-1)
  !
  
  END SUBROUTINE LS_SURFACE_FORCE_3D 
  
  
  SUBROUTINE BC_3D_LS(Nx , Ny ,  NZ ,  U  )
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Routine to compute cell-face pressure on uniform cell-center FV.
  !Compute cell-face velocity using simple average
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    INTEGER,INTENT(IN) :: Nx , Ny  , NZ 
    REAL(KIND=DP),INTENT(INOUT) :: U(:,:,:)
    INTEGER :: I , J , K 


    !2nd order Neumann
    DO K = 2 , NZ-1
    DO J = 2 , Ny-1
    !	U(1,J,K)=U(2,J,K)
    !	U(Nx,J,K)=U(Nx-1,J,K)
      U(1,J,K)  = ( 9.0_DP*U(2,J,K)    - U(3,J,K) ) / 8.0_DP
      U(Nx,J,K) = ( 9.0_DP*U(Nx-1,J,K) - U(Nx-2,J,K) ) / 8.0_DP
     
    END DO
    ENDDO

    DO K = 2  ,NZ-1
    DO I = 2 , Nx-1
    !	U(I,1,K) = U(I,2,K)
    !	U(I,Ny,K) = U(I,Ny-1,K)
      U(I,1,K)  = ( 9.0_DP*U(I,2,K)    - U(I,3,K) ) / 8.0_DP
      U(I,Ny,K) = ( 9.0_DP*U(I,Ny-1,K) - U(I,Ny-2,K) ) / 8.0_DP
      
    ENDDO
    ENDDO
    
    DO J = 2 , NY-1
    DO I = 2 , Nx-1
    !	U(I,J,1) = U(I,J,2)
    !	U(I,J,Nz) = U(I,J,Nz-1)
    	
      U(I,J,1)  = ( 9.0_DP*U(I,J,2)    - U(I,J,3) ) / 8.0_DP
      U(I,J,NZ) = ( 9.0_DP*U(I,J,NZ-1) - U(I,J,NZ-2) ) / 8.0_DP
     
     

      
    ENDDO
    ENDDO
    
    
    U(1,1,1)   =  ( 9.0_DP*U(2,2,2)       - U(3,3,3) ) / 8.0_DP
    U(1,Ny,1)  =  ( 9.0_DP*U(2,Ny-1,2)    - U(3,Ny-2,3) ) / 8.0_DP
    U(Nx,1,1)  =  ( 9.0_DP*U(Nx-1,2,2)    - U(Nx-2,3,3) ) / 8.0_DP
    U(Nx,NY,1)  = ( 9.0_DP*U(Nx-1,NY-1,2)    - U(Nx-2,NY-2,3) ) / 8.0_DP
    
    U(1,1,NZ)   = ( 9.0_DP*U(2,2,NZ-1)       - U(3,3,NZ-2) ) / 8.0_DP
    U(1,Ny,NZ)  = ( 9.0_DP*U(2,Ny-1,NZ-1)    - U(3,Ny-2,NZ-2) ) / 8.0_DP
    U(Nx,1,NZ) =  ( 9.0_DP*U(Nx-1,2,NZ-1) - U(Nx-2,3,NZ-2) ) / 8.0_DP
    U(Nx,NY,NZ)  = ( 9.0_DP*U(Nx-1,NY-1,NZ-1)    - U(Nx-2,NY-2,NZ-2) ) / 8.0_DP

   ! U(1,1)   =  ( 9.0_DP*U(2,2)       - U(3,3) ) / 8.0_DP
   ! U(1,Ny)  =  ( 9.0_DP*U(2,Ny-1)    - U(3,Ny-2) ) / 8.0_DP
   ! U(Nx,1)  =  ( 9.0_DP*U(Nx-1,2)    - U(Nx-2,3) ) / 8.0_DP
   ! U(Nx,Ny) =  ( 9.0_DP*U(Nx-1,Ny-1) - U(Nx-2,Ny-2) ) / 8.0_DP
   
   !  
   !  DO I = 1, Nx-4
   !  DO J = 1, Ny-4
   !  	
   !  IF ( U(I,J,1)*U(I,J+1,1) < 0.0_DP) THEN
   !  	IF ( U(I,J+2,1) > 0.0_DP ) THEN
   !  		IF ( v(I,J+1,1) > 0.0_DP) THEN
   !  			IF
   !  


  END SUBROUTINE BC_3D_LS
  
SUBROUTINE BC_3D_LS1(Nx , Ny, Nz, DZ, U, ANGC, U_x, U_y)
  INTEGER, INTENT(IN) :: Nx,Ny,Nz
  REAL(KIND=DP),INTENT(INOUT) :: U(:,:,:), U_x(:,:,:), U_y(:,:,:)
  INTEGER :: I,J,K
  REAL(KIND=DP),INTENT(IN):: ANGC, Dz
  
  !! HYDROPHOBIC EFFECT
  
  DO I = 2, Nx-1
  DO J = 2, Ny-1
  	
  	
    IF( U_X(I,J,1) >= U_Y(I,J,1) )THEN
    IF( U(I+1,J,1)*U(I,J,1) < 0.0_DP) THEN	
    
    IF( U(I,J,1) > 0.0_DP )THEN  		
      U(I,J,1) = U(I,J,2) + DZ * COS(ANGC)
    ELSEIF( U(I,J,1) < 0.0_DP )THEN
    	U(I+1,J,1) = U(I+1,J,2) + DZ * COS(ANGC)
    ELSE                   
    U(I,J,1)     = U(I,J,2)	
    ENDIF
    ENDIF
    
    ELSEIF( U_Y(I,J,1) > U_X(I,J,1) )THEN
    IF( U(I,J+1,1)*U(I,J,1) < 0.0_DP) THEN	
    IF( U(I,J,1) > 0.0_DP )THEN  		
      U(I,J,1) = U(I,J,2) + DZ * COS(ANGC)
    ELSEIF( U(I,J,1) < 0.0_DP )THEN
    	U(I,J+1,1) = U(I,J+1,2) + DZ * COS(ANGC)
    ELSE                   
      U(I,J,1)     = U(I,J,2)	
    ENDIF
    ENDIF
    ENDIF
 	
  	U(I,J,Nz) = U(I,J,Nz-1)
  	 	
END DO 
END DO
  			
  DO J=2, Ny-1
  DO K=2, Nz-1
  	
  	U(1,J,K) = U(2,J,K)
  	U(Nx,J,K) = U(Nx-1,J,K)
  	
  	END DO
    END DO
    
  
  DO K=2, Nz-1
  DO I=2, Nx-1
  	
  	U(I,1,K) = U(I,2,K)
  	U(I,Ny,K) = U(I,Ny-1,K)
  	
  END DO
  END DO
		
  	 	
  END SUBROUTINE BC_3D_LS1
 
  
   SUBROUTINE BC_2D_P_3D(Nx , Ny , NZ , P)
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !Routine to compute cell-face pressure on uniform cell-center FV.
   !Compute cell-face velocity using simple average
   !
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER,INTENT(IN) :: Nx , Ny , NZ
     REAL(KIND=DP),INTENT(INOUT) :: P(:,:,:)
 
     INTEGER :: I , J , K
 
  !!L
    DO K = 2 , NZ-1
    DO J = 2 , Ny-1
   
   
      P(1,J,K)  =  2.0_DP*P(2,J,K)    - P(3,J,K) 
      P(Nx,J,K) =  2.0_DP*P(Nx-1,J,K) - P(Nx-2,J,K) 
    ENDDO
    ENDDO

    DO K = 2  ,NZ-1
    DO I = 2 , Nx-1
  
  
    	
      P(I,1,K)  =  2.0_DP*P(I,2,K)    - P(I,3,K) 
      P(I,Ny,K) =  2.0_DP*P(I,Ny-1,K) - P(I,Ny-2,K) 
      
    ENDDO
    ENDDO
    
    DO J = 2  ,NY-1
    DO I = 2 , Nx-1
    	
    	
    	
      P(I,J,1)  =  2.0_DP*P(I,J,2)    - P(I,J,3) 
      P(I,J,NZ) =  2.0_DP*P(I,J,NZ-1) - P(I,J,NZ-2)
      
    ENDDO
    ENDDO
    
    P(1,1,1)   =   2.0_DP*P(2,2,2)       - P(3,3,3) 
    P(1,Ny,1)  =   2.0_DP*P(2,Ny-1,2)    - P(3,Ny-2,3) 
    P(Nx,1,1)  =   2.0_DP*P(Nx-1,2,2)    - P(Nx-2,3,3)
    P(Nx,NY,1)  =  2.0_DP*P(Nx-1,NY-1,2)  - P(Nx-2,NY-2,3) 
    
    P(1,1,NZ)   =  2.0_DP*P(2,2,NZ-1)       - P(3,3,NZ-2) 
    P(1,Ny,NZ)  =  2.0_DP*P(2,Ny-1,NZ-1)    - P(3,Ny-2,NZ-2) 
    P(Nx,1,NZ)  =  2.0_DP*P(Nx-1,2,NZ-1)    - P(Nx-2,3,NZ-2) 
    P(Nx,NY,NZ) =  2.0_DP*P(Nx-1,NY-1,NZ-1)  - P(Nx-2,NY-2,NZ-2) 
    
 
   DO K = 2 , NZ-1
   DO J = 2 , Ny-1
   	
      P(1,J,K)  = (P(1,J,K)    + P(2,J,K)  ) * 0.5_DP
      P(Nx,J,K) = (P(Nx,J,K) + P(Nx-1,J,K) ) * 0.5_DP
      
   END DO
   END DO

   DO K = 2  ,NZ-1
   DO I = 2 , Nx-1
   	
      P(I,1,K)  = (P(I,1,K)    + P(I,2,K)  ) * 0.5_DP
      P(I,Ny,K) = (P(I,Ny,K) + P(I,Ny-1,K) ) * 0.5_DP
      
   END DO
   END DO
   
   
   DO J = 2  ,NY-1
   DO I = 2 , Nx-1
   	
      P(I,J,1)  = (P(I,J,1)    + P(I,J,2)  ) * 0.5_DP
      P(I,J,NZ) = (P(I,J,NZ) + P(I,J,NZ-1) ) * 0.5_DP
      
   END DO
   END DO
   

    P(1,1,1)   =   P(2,2,2)      
    P(1,Ny,1)  =   P(2,Ny-1,2)    
    P(Nx,1,1)  =   P(Nx-1,2,2)    
    P(Nx,NY,1)  =  P(Nx-1,NY-1,2)  
    
    P(1,1,NZ)   =  P(2,2,NZ-1)      
    P(1,Ny,NZ)  =  P(2,Ny-1,NZ-1)    
    P(Nx,1,NZ)  =  P(Nx-1,2,NZ-1)    
    P(Nx,NY,NZ) =  P(Nx-1,NY-1,NZ-1) 
     
     
   !  P(1,1)  = (P(1,2) + P(2,1)) * 0.5_DP
   !  P(Nx,1) = (P(Nx-1,1) + P(Nx,2)) * 0.5_DP
   ! 
   !  P(1,Ny)  = (P(2,Ny) + P(1,Ny-1)) * 0.5_DP
   !  P(Nx,Ny) = (P(Nx-1,Ny) + P(Nx,Ny-1)) * 0.5_DP
   
 
   END SUBROUTINE BC_2D_P_3D
  
  
  
   SUBROUTINE RHS_COMPUTE_QUICK3ST_3D_1(Nx , Ny , NZ , Dx , Dy , DZ , C_X , C_Y  , C_Z   &
                                &, K_X , K_Y , K_Z , RHO , Mu , S , U , RHS  )
                               
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
     !Routine to compute RHS for using DRP for convective terms.
     !For the diffusion terms, CD2 is used.
     !
     !
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       IMPLICIT NONE
    
       INTEGER      ,INTENT(IN) :: Nx , Ny , NZ
       REAL(KIND=DP),INTENT(IN) :: Dx , Dy , DZ
    
       REAL(KIND=DP),INTENT(IN) :: C_X(:,:,:) , C_Y(:,:,:) , C_Z(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: K_X(:,:,:) , K_Y(:,:,:) , K_Z(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: RHO(:,:,:) , Mu(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: S(:,:,:) , U(:,:,:)
    
       REAL(KIND=DP),INTENT(OUT) :: RHS(:,:,:)
    
       REAL(KIND=DP) :: U_X(Nx,Ny,NZ) , U_Y(Nx,Ny,NZ) , U_old(0:Nx+1,0:Ny+1,0:NZ+1)
       REAL(KIND=DP) :: U_XX(Nx,Ny,NZ) , U_YY(Nx,Ny,NZ) , U_ZZ(Nx,Ny,NZ) 
       REAL(KIND=DP) :: U_XXX(Nx,Ny,NZ) , U_YYY(Nx,Ny,NZ) , U_ZZZ(Nx,Ny,NZ)
       REAL(KIND=DP) :: K_Xc(Nx,Ny,NZ) , K_Yc(Nx,Ny,NZ) , K_Zc(Nx,Ny,NZ)
       REAL(KIND=DP) :: UC_X(Nx,Ny,NZ) , UC_Y(Nx,Ny,NZ) , UC_Z(Nx,Ny,NZ)
       INTEGER :: I , J , K 
      
     
       REAL(KIND=DP) :: CX_E(Nx,Ny,NZ)  , CY_N(Nx,Ny,NZ) ,  CZ_C(Nx,Ny,NZ) 
       REAL(KIND=DP) :: LIN_E(Nx,Ny,NZ) , LIN_N(Nx,Ny,NZ) , LIN_C(Nx,Ny,NZ)
       REAL(KIND=DP) :: CURV_E(Nx,Ny,NZ)  , CURV_N(Nx,Ny,NZ) , CURV_C(Nx,Ny,NZ) 
       REAL(KIND=DP) :: U_E(Nx,Ny,NZ) , U_N(Nx,Ny,NZ) , U_C(Nx,Ny,NZ)
       REAL(KIND=DP) :: TEMP
      
       DO K = 1 , NZ
       DO J = 1 , Ny
       DO I = 1 , Nx
         U_old(I,J,K) = U(I,J,K)
       END DO
       END DO
       ENDDO
     
       DO K = 1 , NZ
       DO J = 1 , Ny
         U_old(1,J,K)  = ( 8.0_DP * U(1,J,K)  - 6.0_DP * U(2,J,K)    + U(3,J,K)    ) / 3.0_DP
         U_old(Nx,J,K) = ( 8.0_DP * U(Nx,J,K) - 6.0_DP * U(Nx-1,J,K) + U(Nx-2,J,K) ) / 3.0_DP
    
         U_old(0,J,K)  =   U_OLD(1,J,K)  
         U_old(Nx+1,J,K) = U_OLD(Nx,J,K) 
    
       END DO
       END DO
         
       DO K = 1  ,NZ
       DO I = 1 , Nx
         U_old(I,1,K)  = ( 8.0_DP * U(I,1,K)  - 6.0_DP * U(I,2,K)    + U(I,3,K)    ) / 3.0_DP
         U_old(I,Ny,K) = ( 8.0_DP * U(I,Ny,K) - 6.0_DP * U(I,Ny-1,K) + U(I,Ny-2,K) ) / 3.0_DP
    
         U_old(I,0,K)  = U_OLD(I,1,K)  
         U_old(I,Ny+1,K) = U_OLD(I,Ny,K) 
       END DO
       END DO
       
          
       DO J = 1 , NY
       DO I = 1 , Nx
         U_old(I,J,1)  = ( 8.0_DP * U(I,J,1)  - 6.0_DP * U(I,J,2)    + U(I,J,3)    ) / 3.0_DP
         U_old(I,J,NZ) = ( 8.0_DP * U(I,J,NZ) - 6.0_DP * U(I,J,NZ-1) + U(I,J,NZ-2) ) / 3.0_DP 
         
         U_old(I,J,0)  = U_OLD(I,J,1)  
         U_old(I,J,NZ+1) = U_OLD(I,J,NZ)    
       END DO
       END DO          
       
       U_old(1,1,1)   = (8.0_DP * U(1,1,1)   - 6.0_DP * U(2,2,2)       + U(3,3,3)      ) / 3.0_DP
       U_old(1,Ny,1)  = (8.0_DP * U(1,Ny,1)  - 6.0_DP * U(2,Ny-1,2)    + U(3,Ny-2,3)   ) / 3.0_DP
       U_old(Nx,1,1)  = (8.0_DP * U(Nx,1,1)  - 6.0_DP * U(Nx-1,2,2)    + U(Nx-2,3,3)   ) / 3.0_DP
       U_old(Nx,Ny,1) = (8.0_DP * U(Nx,Ny,1) - 6.0_DP * U(Nx-1,Ny-1,2) + U(Nx-2,Ny-2,3)) / 3.0_DP
    
       U_old(1,1,NZ)   = (8.0_DP * U(1,1,NZ)   - 6.0_DP * U(2,2,NZ-1)       + U(3,3,NZ-2)      ) / 3.0_DP
       U_old(1,Ny,NZ)  = (8.0_DP * U(1,Ny,NZ)  - 6.0_DP * U(2,Ny-1,NZ-1)    + U(3,Ny-2,NZ-2)   ) / 3.0_DP
       U_old(Nx,1,NZ)  = (8.0_DP * U(Nx,1,NZ)  - 6.0_DP * U(Nx-1,2,NZ-1)    + U(Nx-2,3,NZ-2)   ) / 3.0_DP
       U_old(Nx,Ny,NZ) = (8.0_DP * U(Nx,Ny,NZ) - 6.0_DP * U(Nx-1,Ny-1,NZ-1) + U(Nx-2,Ny-2,NZ-2)) / 3.0_DP
    
    
  
       !For diffusion term
         !For cell face K_X and K_Y
       CALL Cell_face_velocity_FV_3D(Nx , Ny , NZ ,  Mu , Mu  , Mu  , K_Xc , K_Yc , K_Zc)    
       
      
         !for cell face derivative
      ! CALL Cell_face_derivative_FV(Nx , Ny , Dx , Dy , U , U_XX , U_YY)
       
       CALL Cell_face_derivative_FV_3D(Nx , Ny ,  NZ , Dx , Dy  , DZ , U , U_XX , U_YY , U_ZZ)  
       
       

    
      DO K = 2  , NZ-1 
      DO J = 2  , NY-1 
       	 CX_E(1,J,K) =  C_X(1,J,K) 
       	 CX_E(NX-1,J,K) =  C_X(NX,J,K)     	 
      ENDDO
      ENDDO
      
       DO K = 2  ,NZ -1 
       DO J = 2  ,NY -1 
       DO I = 2 , NX -2
       	 CX_E(I,J,K) = 0.5_DP* ( C_X(I+1,J,K) + C_X(I,J,K) )
       ENDDO
       ENDDO	
       ENDDO
        
        
       DO K = 2 , NZ-1
       DO I = 2 , NX-1
         CY_N(I,1,K)    = C_Y(I,1,K)
         CY_N(I,Ny-1,K) = C_Y(I,Ny,K)
       END DO
       END DO
       
  
       DO K = 2 , NZ-1
       DO J = 2 , Ny-2
       DO I = 2 , Nx-1
          CY_N(I,J,K) = 0.5_DP * ( C_Y(I,J+1,K) + C_Y(I,J,K) )
       END DO
       END DO
       END DO
       
       
       DO J = 2 , NY-1
       DO I = 2 , NX-1
         CZ_C(I,J,1)    = C_Z(I,J,1)
         CZ_C(I,J,NZ-1) = C_Z(I,J,NZ)
       END DO
       END DO
       
  
       DO K = 2 , NZ-2
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
          CZ_C(I,J,K) = 0.5_DP * ( C_Z(I,J,K+1) + C_Z(I,J,K) )
       END DO
       END DO
       END DO
       
       
       
      
     
       DO K = 1  ,NZ -1
       DO J = 1  ,NY -1 
       DO I = 1 , NX -1
       	
         IF(CX_E(I,J,K)>=0.0_DP) THEN
       	   U_E(I,J,K) =   1.0_DP / 8.0_DP*( -U_old(I-1,J,K) + 6.0_DP*U_OLD(I,J,K) +  3.0_DP*U_OLD(I+1,J,K)  )
       	 ELSE  
       	   U_E(I,J,K) =   1.0_DP / 8.0_DP*( -U_old(I+2,J,K) + 6.0_DP*U_OLD(I+1,J,K) +  3.0_DP*U_OLD(I,J,K)  )
       	 ENDIF  
       	   
       	 
       ENDDO
       ENDDO  
       ENDDO
       
       
     
       
       
       DO K = 1  ,NZ -1
       DO J = 1  ,NY -1 
       DO I = 1 , NX -1
         IF(CY_N(I,J,K)>=0.0_DP) THEN
       	   U_N(I,J,K) =   1.0_DP / 8.0_DP*( -U_old(I,J-1,K) + 6.0_DP*U_OLD(I,J,K) +  3.0_DP*U_OLD(I,J+1,K)  )
       	 ELSE  
       	   U_N(I,J,K) =   1.0_DP / 8.0_DP*( -U_old(I,J+2,K) + 6.0_DP*U_OLD(I,J+1,K) +  3.0_DP*U_OLD(I,J,K)  )
       	 ENDIF  
       	 
       ENDDO
       ENDDO  
       ENDDO
       
       
       DO K = 1  ,NZ -1
       DO J = 1  ,NY -1 
       DO I = 1 , NX -1
         
       	IF( CZ_C(I,J,K)>=0.0_DP) THEN  
       	   U_C(I,J,K) =   1.0_DP / 8.0_DP*( -U_old(I,J,K-1) + 6.0_DP*U_OLD(I,J,K) +  3.0_DP*U_OLD(I,J,K+1)  )
       	ELSE
       	   U_C(I,J,K) =   1.0_DP / 8.0_DP*( -U_old(I,J,K+2) + 6.0_DP*U_OLD(I,J,K+1) +  3.0_DP*U_OLD(I,J,K)  )
       	ENDIF
       	   
       ENDDO
       ENDDO  
       ENDDO
       
  
       
            
     !  !Compute RHS
     !  !convective terms and source term
  
  
     

       DO K = 2 , NZ-1
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
         RHS(I,J,K) = -(  CX_E(I,J,K)*U_E(I,J,K) - CX_E(I-1,J,K)*U_E(I-1,J,K)  ) / DX    & 
                    & -(  CY_N(I,J,K)*U_N(I,J,K) - CY_N(I,J-1,K)*U_N(I,J-1,K)  ) / DY    &
                    & -(  CZ_C(I,J,K)*U_C(I,J,K) - CZ_C(I,J,K-1)*U_C(I,J,K-1)  ) / DZ       + S(I,J,K)
       END DO
       END DO
       END DO
  
 
    
       !diffusion terms
       DO K = 2 , NZ-1
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
         RHS(I,J,K) = RHS(I,J,K) + ( K_X(I,J,K)*(U_XX(I,J,K) * K_Xc(I,J,K) - U_XX(I-1,J,K) * K_Xc(I-1,J,K) ) / Dx &
                  &                 +K_Y(I,J,K)*(U_YY(I,J,K) * K_Yc(I,J,K) - U_YY(I,J-1,K) * K_Yc(I,J-1,K) ) / Dy &
                  &                 +K_Z(I,J,K)*(U_ZZ(I,J,K) * K_Zc(I,J,K) - U_ZZ(I,J,K-1) * K_Zc(I,J,K-1) ) / DZ &                 
                  &             ) / RHO(I,J,K)
       END DO
       END DO
       END DO
    
    
     END SUBROUTINE RHS_COMPUTE_QUICK3ST_3D_1
     
     
     SUBROUTINE RHS_COMPUTE_QUICK3ST_3D(Nx , Ny , NZ , Dx , Dy , DZ , C_X , C_Y  , C_Z   &
                                &, K_X , K_Y , K_Z , RHO , Mu , S , U , RHS  )
                               
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
     !Routine to compute RHS for using DRP for convective terms.
     !For the diffusion terms, CD2 is used.
     !
     !
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       IMPLICIT NONE
    
       INTEGER      ,INTENT(IN) :: Nx , Ny , NZ
       REAL(KIND=DP),INTENT(IN) :: Dx , Dy , DZ
    
       REAL(KIND=DP),INTENT(IN) :: C_X(:,:,:) , C_Y(:,:,:) , C_Z(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: K_X(:,:,:) , K_Y(:,:,:) , K_Z(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: RHO(:,:,:) , Mu(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: S(:,:,:) , U(:,:,:)
    
       REAL(KIND=DP),INTENT(OUT) :: RHS(:,:,:)
    
       REAL(KIND=DP) :: U_X(Nx,Ny,NZ) , U_Y(Nx,Ny,NZ) , U_old(0:Nx+1,0:Ny+1,0:NZ+1)
       REAL(KIND=DP) :: U_XX(Nx,Ny,NZ) , U_YY(Nx,Ny,NZ) , U_ZZ(Nx,Ny,NZ) 
       REAL(KIND=DP) :: U_XXX(Nx,Ny,NZ) , U_YYY(Nx,Ny,NZ) , U_ZZZ(Nx,Ny,NZ)
       REAL(KIND=DP) :: K_Xc(Nx,Ny,NZ) , K_Yc(Nx,Ny,NZ) , K_Zc(Nx,Ny,NZ)
       REAL(KIND=DP) :: UC_X(Nx,Ny,NZ) , UC_Y(Nx,Ny,NZ) , UC_Z(Nx,Ny,NZ)
       INTEGER :: I , J , K 
      
     
       REAL(KIND=DP) :: CX_E(Nx,Ny,NZ)  , CY_N(Nx,Ny,NZ) ,  CZ_C(Nx,Ny,NZ) 
       REAL(KIND=DP) :: LIN_E(Nx,Ny,NZ) , LIN_N(Nx,Ny,NZ) , LIN_C(Nx,Ny,NZ)
       REAL(KIND=DP) :: CURV_E(Nx,Ny,NZ)  , CURV_N(Nx,Ny,NZ) , CURV_C(Nx,Ny,NZ) 
       REAL(KIND=DP) :: U_E(Nx,Ny,NZ) , U_N(Nx,Ny,NZ) , U_C(Nx,Ny,NZ)
       REAL(KIND=DP) :: TEMP
      
       DO K = 1 , NZ
       DO J = 1 , Ny
       DO I = 1 , Nx
         U_old(I,J,K) = U(I,J,K)
       END DO
       END DO
       ENDDO
     
       DO K = 1 , NZ
       DO J = 1 , Ny
         U_old(1,J,K)  = ( 8.0_DP * U(1,J,K)  - 6.0_DP * U(2,J,K)    + U(3,J,K)    ) / 3.0_DP
         U_old(Nx,J,K) = ( 8.0_DP * U(Nx,J,K) - 6.0_DP * U(Nx-1,J,K) + U(Nx-2,J,K) ) / 3.0_DP
    
         U_old(0,J,K)  =   U_OLD(1,J,K)  
         U_old(Nx+1,J,K) = U_OLD(Nx,J,K) 
    
       END DO
       END DO
         
       DO K = 1  ,NZ
       DO I = 1 , Nx
         U_old(I,1,K)  = ( 8.0_DP * U(I,1,K)  - 6.0_DP * U(I,2,K)    + U(I,3,K)    ) / 3.0_DP
         U_old(I,Ny,K) = ( 8.0_DP * U(I,Ny,K) - 6.0_DP * U(I,Ny-1,K) + U(I,Ny-2,K) ) / 3.0_DP
    
         U_old(I,0,K)  = U_OLD(I,1,K)  
         U_old(I,Ny+1,K) = U_OLD(I,Ny,K) 
       END DO
       END DO
       
          
       DO J = 1 , NY
       DO I = 1 , Nx
         U_old(I,J,1)  = ( 8.0_DP * U(I,J,1)  - 6.0_DP * U(I,J,2)    + U(I,J,3)    ) / 3.0_DP
         U_old(I,J,NZ) = ( 8.0_DP * U(I,J,NZ) - 6.0_DP * U(I,J,NZ-1) + U(I,J,NZ-2) ) / 3.0_DP 
         
         U_old(I,J,0)  = U_OLD(I,J,1)  
         U_old(I,J,NZ+1) = U_OLD(I,J,NZ)    
       END DO
       END DO          
       
       U_old(1,1,1)   = (8.0_DP * U(1,1,1)   - 6.0_DP * U(2,2,2)       + U(3,3,3)      ) / 3.0_DP
       U_old(1,Ny,1)  = (8.0_DP * U(1,Ny,1)  - 6.0_DP * U(2,Ny-1,2)    + U(3,Ny-2,3)   ) / 3.0_DP
       U_old(Nx,1,1)  = (8.0_DP * U(Nx,1,1)  - 6.0_DP * U(Nx-1,2,2)    + U(Nx-2,3,3)   ) / 3.0_DP
       U_old(Nx,Ny,1) = (8.0_DP * U(Nx,Ny,1) - 6.0_DP * U(Nx-1,Ny-1,2) + U(Nx-2,Ny-2,3)) / 3.0_DP
    
       U_old(1,1,NZ)   = (8.0_DP * U(1,1,NZ)   - 6.0_DP * U(2,2,NZ-1)       + U(3,3,NZ-2)      ) / 3.0_DP
       U_old(1,Ny,NZ)  = (8.0_DP * U(1,Ny,NZ)  - 6.0_DP * U(2,Ny-1,NZ-1)    + U(3,Ny-2,NZ-2)   ) / 3.0_DP
       U_old(Nx,1,NZ)  = (8.0_DP * U(Nx,1,NZ)  - 6.0_DP * U(Nx-1,2,NZ-1)    + U(Nx-2,3,NZ-2)   ) / 3.0_DP
       U_old(Nx,Ny,NZ) = (8.0_DP * U(Nx,Ny,NZ) - 6.0_DP * U(Nx-1,Ny-1,NZ-1) + U(Nx-2,Ny-2,NZ-2)) / 3.0_DP
    
    
  
       !For diffusion term
         !For cell face K_X and K_Y
       CALL Cell_face_velocity_FV_3D(Nx , Ny , NZ ,  Mu , Mu  , Mu  , K_Xc , K_Yc , K_Zc)    
       
      
         !for cell face derivative
      ! CALL Cell_face_derivative_FV(Nx , Ny , Dx , Dy , U , U_XX , U_YY)
       
       CALL Cell_face_derivative_FV_3D(Nx , Ny ,  NZ , Dx , Dy  , DZ , U , U_XX , U_YY , U_ZZ)  
       
       

    
      DO K = 2  , NZ-1 
      DO J = 2  , NY-1 
       	 CX_E(1,J,K) =  C_X(1,J,K) 
       	 CX_E(NX-1,J,K) =  C_X(NX,J,K)     	 
      ENDDO
      ENDDO
      
       DO K = 2  ,NZ -1 
       DO J = 2  ,NY -1 
       DO I = 2 , NX -2
       	 CX_E(I,J,K) = 0.5_DP* ( C_X(I+1,J,K) + C_X(I,J,K) )
       ENDDO
       ENDDO	
       ENDDO
        
        
       DO K = 2 , NZ-1
       DO I = 2 , NX-1
         CY_N(I,1,K)    = C_Y(I,1,K)
         CY_N(I,Ny-1,K) = C_Y(I,Ny,K)
       END DO
       END DO
       
  
       DO K = 2 , NZ-1
       DO J = 2 , Ny-2
       DO I = 2 , Nx-1
          CY_N(I,J,K) = 0.5_DP * ( C_Y(I,J+1,K) + C_Y(I,J,K) )
       END DO
       END DO
       END DO
       
       
       DO J = 2 , NY-1
       DO I = 2 , NX-1
         CZ_C(I,J,1)    = C_Z(I,J,1)
         CZ_C(I,J,NZ-1) = C_Z(I,J,NZ)
       END DO
       END DO
       
  
       DO K = 2 , NZ-2
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
          CZ_C(I,J,K) = 0.5_DP * ( C_Z(I,J,K+1) + C_Z(I,J,K) )
       END DO
       END DO
       END DO
       
       
       
      
     
   DO K = 1  ,NZ -1 
   DO J = 1  ,NY -1 
   DO I = 1 , NX -1
   	   
   	  TEMP = CX_E(I,J,K)
      LIN_E(I,J,K) = 0.5_dP*( U_OLD(I+1,J,K) + U_OLD(I,J,K)  )
      CURV_E(I,J,K) = 0.5_DP*( 1.0_DP + SIGN(1.0_DP,TEMP) ) * ( U_OLD(I+1,J,K)-2.0_DP*U_OLD(I,J,K)+ U_OLD(I-1,J,K) )  &
              &   + 0.5_DP*( 1.0_DP - SIGN(1.0_DP,TEMP) ) * ( U_OLD(I+2,J,K)-2.0_DP*U_OLD(I+1,J,K)+ U_OLD(I,J,K) )
              
      TEMP = CY_N(I,J,K)
      LIN_N(I,J,K) = 0.5_dP*( U_OLD(I,J+1,K) + U_OLD(I,J,K)  )
      CURV_N(I,J,K) = 0.5_DP*( 1.0_DP + SIGN(1.0_DP,TEMP) ) * ( U_OLD(I,J+1,K)-2.0_DP*U_OLD(I,J,K)+ U_OLD(I,J-1,K) )  &
              &   + 0.5_DP*( 1.0_DP - SIGN(1.0_DP,TEMP) ) * ( U_OLD(I,J+2,K)-2.0_DP*U_OLD(I,J+1,K)+ U_OLD(I,J,K) ) 
              
      TEMP = CZ_C(I,J,K)
      LIN_C(I,J,K) = 0.5_dP*( U_OLD(I,J,K+1) + U_OLD(I,J,K)  )
      CURV_C(I,J,K) = 0.5_DP*( 1.0_DP + SIGN(1.0_DP,TEMP) ) * ( U_OLD(I,J,K+1)-2.0_DP*U_OLD(I,J,K)+ U_OLD(I,J,K-1) )  &
              &   + 0.5_DP*( 1.0_DP - SIGN(1.0_DP,TEMP) ) * ( U_OLD(I,J,K+2)-2.0_DP*U_OLD(I,J,K+1)+ U_OLD(I,J,K) )                
               
   ENDDO
   ENDDO
   ENDDO
     
    
    
    
     
       DO K = 1  ,NZ -1
       DO J = 1  ,NY -1 
       DO I = 1 , NX -1
       	   U_E(I,J,K) =   LIN_E(I,J,K) - 1.0_DP / 8.0_DP*CURV_E(I,J,K)
       	   U_N(I,J,K) =   LIN_N(I,J,K) - 1.0_DP / 8.0_DP*CURV_N(I,J,K)
       	   U_C(I,J,K) =   LIN_C(I,J,K) - 1.0_DP / 8.0_DP*CURV_C(I,J,K)
       ENDDO
       ENDDO  
       ENDDO
       
  
       
            
     !  !Compute RHS
     !  !convective terms and source term
  
  
     

       DO K = 2 , NZ-1
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
         RHS(I,J,K) = -(  CX_E(I,J,K)*U_E(I,J,K) - CX_E(I-1,J,K)*U_E(I-1,J,K)  ) / DX    & 
                    & -(  CY_N(I,J,K)*U_N(I,J,K) - CY_N(I,J-1,K)*U_N(I,J-1,K)  ) / DY    &
                    & -(  CZ_C(I,J,K)*U_C(I,J,K) - CZ_C(I,J,K-1)*U_C(I,J,K-1)  ) / DZ       + S(I,J,K)
       END DO
       END DO
       END DO
  
 
    
       !diffusion terms
       DO K = 2 , NZ-1
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
         RHS(I,J,K) = RHS(I,J,K) + ( K_X(I,J,K)*(U_XX(I,J,K) * K_Xc(I,J,K) - U_XX(I-1,J,K) * K_Xc(I-1,J,K) ) / Dx &
                  &                 +K_Y(I,J,K)*(U_YY(I,J,K) * K_Yc(I,J,K) - U_YY(I,J-1,K) * K_Yc(I,J-1,K) ) / Dy &
                  &                 +K_Z(I,J,K)*(U_ZZ(I,J,K) * K_Zc(I,J,K) - U_ZZ(I,J,K-1) * K_Zc(I,J,K-1) ) / DZ &                 
                  &             ) / RHO(I,J,K)
       END DO
       END DO
       END DO
    
    
     END SUBROUTINE RHS_COMPUTE_QUICK3ST_3D
     
     
      SUBROUTINE RHS_COMPUTE_UP1ST(Nx , Ny , NZ , Dx , Dy , DZ , C_X , C_Y  , C_Z   &
                                &, K_X , K_Y , K_Z , RHO , Mu , S , U , RHS  )
                               
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!     
     !Routine to compute RHS for using DRP for convective terms.
     !For the diffusion terms, CD2 is used.
     !
     !
     !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
       IMPLICIT NONE
    
       INTEGER      ,INTENT(IN) :: Nx , Ny , NZ
       REAL(KIND=DP),INTENT(IN) :: Dx , Dy , DZ
    
       REAL(KIND=DP),INTENT(IN) :: C_X(:,:,:) , C_Y(:,:,:) , C_Z(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: K_X(:,:,:) , K_Y(:,:,:) , K_Z(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: RHO(:,:,:) , Mu(:,:,:)
       REAL(KIND=DP),INTENT(IN) :: S(:,:,:) , U(:,:,:)
    
       REAL(KIND=DP),INTENT(OUT) :: RHS(:,:,:)
    
       REAL(KIND=DP) :: U_X(Nx,Ny,NZ) , U_Y(Nx,Ny,NZ) , U_Z(Nx,Ny,NZ) , U_old(0:Nx+1,0:Ny+1,0:NZ+1)
       REAL(KIND=DP) :: U_XX(Nx,Ny,NZ) , U_YY(Nx,Ny,NZ) , U_ZZ(Nx,Ny,NZ) 
       REAL(KIND=DP) :: U_XXX(Nx,Ny,NZ) , U_YYY(Nx,Ny,NZ) , U_ZZZ(Nx,Ny,NZ)
       REAL(KIND=DP) :: K_Xc(Nx,Ny,NZ) , K_Yc(Nx,Ny,NZ) , K_Zc(Nx,Ny,NZ)
       REAL(KIND=DP) :: UC_X(Nx,Ny,NZ) , UC_Y(Nx,Ny,NZ) , UC_Z(Nx,Ny,NZ)
     !  REAL(KIND=DP) :: UC_X(Nx,Ny,NZ) ,  UC_X(Nx,Ny,NZ) , UC_X(Nx,Ny,NZ)
       
       INTEGER :: I , J , K 
      
     
       REAL(KIND=DP) :: CX_E(Nx,Ny,NZ)  , CY_N(Nx,Ny,NZ) ,  CZ_C(Nx,Ny,NZ) 
       REAL(KIND=DP) :: LIN_E(Nx,Ny,NZ) , LIN_N(Nx,Ny,NZ) , LIN_C(Nx,Ny,NZ)
       REAL(KIND=DP) :: CURV_E(Nx,Ny,NZ)  , CURV_N(Nx,Ny,NZ) , CURV_C(Nx,Ny,NZ) 
       REAL(KIND=DP) :: U_E(Nx,Ny,NZ) , U_N(Nx,Ny,NZ) , U_C(Nx,Ny,NZ)
       REAL(KIND=DP) :: TEMP
      
       DO K = 1 , NZ
       DO J = 1 , Ny
       DO I = 1 , Nx
         U_old(I,J,K) = U(I,J,K)
       END DO
       END DO
       ENDDO
     
       DO K = 1 , NZ
       DO J = 1 , Ny
         U_old(1,J,K)  = U_old(2,J,K)
         U_old(Nx,J,K) = U_old(Nx-1,J,K)
    
         U_old(0,J,K)  =   U_OLD(1,J,K)  
         U_old(Nx+1,J,K) = U_OLD(Nx,J,K) 
    
       END DO
       END DO
         
       DO K = 1  ,NZ
       DO I = 1 , Nx
         U_old(I,1,K)  =  U_old(I,2,K)
         U_old(I,Ny,K) =  U_old(I,Ny-1,K)
    
         U_old(I,0,K)  = U_OLD(I,1,K)  
         U_old(I,Ny+1,K) = U_OLD(I,Ny,K) 
       END DO
       END DO
       
          
       DO J = 1 , NY
       DO I = 1 , Nx
         U_old(I,J,1)  = U_old(I,J,2)
         U_old(I,J,NZ) =  U_old(I,J,NZ-1)
         
         U_old(I,J,0)  = U_OLD(I,J,1)  
         U_old(I,J,NZ+1) = U_OLD(I,J,NZ)    
       END DO
       END DO          
       
       U_old(1,1,1)   = (8.0_DP * U(1,1,1)   - 6.0_DP * U(2,2,2)       + U(3,3,3)      ) / 3.0_DP
       U_old(1,Ny,1)  = (8.0_DP * U(1,Ny,1)  - 6.0_DP * U(2,Ny-1,2)    + U(3,Ny-2,3)   ) / 3.0_DP
       U_old(Nx,1,1)  = (8.0_DP * U(Nx,1,1)  - 6.0_DP * U(Nx-1,2,2)    + U(Nx-2,3,3)   ) / 3.0_DP
       U_old(Nx,Ny,1) = (8.0_DP * U(Nx,Ny,1) - 6.0_DP * U(Nx-1,Ny-1,2) + U(Nx-2,Ny-2,3)) / 3.0_DP
    
       U_old(1,1,NZ)   = (8.0_DP * U(1,1,NZ)   - 6.0_DP * U(2,2,NZ-1)       + U(3,3,NZ-2)      ) / 3.0_DP
       U_old(1,Ny,NZ)  = (8.0_DP * U(1,Ny,NZ)  - 6.0_DP * U(2,Ny-1,NZ-1)    + U(3,Ny-2,NZ-2)   ) / 3.0_DP
       U_old(Nx,1,NZ)  = (8.0_DP * U(Nx,1,NZ)  - 6.0_DP * U(Nx-1,2,NZ-1)    + U(Nx-2,3,NZ-2)   ) / 3.0_DP
       U_old(Nx,Ny,NZ) = (8.0_DP * U(Nx,Ny,NZ) - 6.0_DP * U(Nx-1,Ny-1,NZ-1) + U(Nx-2,Ny-2,NZ-2)) / 3.0_DP
    
    
  
       !For diffusion term
         !For cell face K_X and K_Y
       CALL Cell_face_velocity_FV_3D(Nx , Ny , NZ ,  Mu , Mu  , Mu  , K_Xc , K_Yc , K_Zc)    
       
      
         !for cell face derivative
      ! CALL Cell_face_derivative_FV(Nx , Ny , Dx , Dy , U , U_XX , U_YY)
       
       CALL Cell_face_derivative_FV_3D(Nx , Ny ,  NZ , Dx , Dy  , DZ , U , U_XX , U_YY , U_ZZ)  
       
       

    
    
     
    
    
    
     
       DO K = 2  ,NZ -1
       DO J = 2  ,NY -1 
       DO I = 2 , NX -1
       	  IF(C_X(I,J,K) >= 0.0_DP) THEN
       	    U_X(I,J,K) = ( U_OLD(I,J,K) -  U_OLD(I-1,J,K) ) /DX 
       	  ELSE 
       	    U_X(I,J,K) = ( U_OLD(I+1,J,K) -  U_OLD(I,J,K) ) /DX
       	  ENDIF 
       	  
       	  IF(C_Y(I,J,K) >= 0.0_DP) THEN
       	    U_Y(I,J,K) = ( U_OLD(I,J,K) -  U_OLD(I,J-1,K) ) /DY 
       	  ELSE 
       	    U_Y(I,J,K) = ( U_OLD(I,J+1,K) -  U_OLD(I,J,K) ) /DY
       	  ENDIF 
       	  
       	  IF(C_Z(I,J,K) >= 0.0_DP) THEN
       	    U_Z(I,J,K) = ( U_OLD(I,J,K) -  U_OLD(I,J,K-1) ) /DZ 
       	  ELSE 
       	    U_Z(I,J,K) = ( U_OLD(I,J,K+1) -  U_OLD(I,J,K) ) /DZ
       	  ENDIF 
       	   
       ENDDO
       ENDDO  
       ENDDO
       
  
       
            
     !  !Compute RHS
     !  !convective terms and source term
  
  
     

       DO K = 2 , NZ-1
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
         RHS(I,J,K) = - C_X(I,J,K)*U_X(I,J,K) -  C_Y(I,J,K)*U_Y(I,J,K) - C_Z(I,J,K)*U_Z(I,J,K)  + S(I,J,K)                  
       END DO
       END DO
       END DO
  
 
    
       !diffusion terms
       DO K = 2 , NZ-1
       DO J = 2 , Ny-1
       DO I = 2 , Nx-1
         RHS(I,J,K) = RHS(I,J,K) + ( K_X(I,J,K)*(U_XX(I,J,K) * K_Xc(I,J,K) - U_XX(I-1,J,K) * K_Xc(I-1,J,K) ) / Dx &
                  &                 +K_Y(I,J,K)*(U_YY(I,J,K) * K_Yc(I,J,K) - U_YY(I,J-1,K) * K_Yc(I,J-1,K) ) / Dy &
                  &                 +K_Z(I,J,K)*(U_ZZ(I,J,K) * K_Zc(I,J,K) - U_ZZ(I,J,K-1) * K_Zc(I,J,K-1) ) / DZ &                 
                  &             ) / RHO(I,J,K)
       END DO
       END DO
       END DO
    
    
     END SUBROUTINE RHS_COMPUTE_UP1ST
     
     
   SUBROUTINE cell_face_POISSON_SOLVER_GS(Nx , Ny , NZ , Dx , Dy , DZ , Dt , Uc , Vc , Wc &
                                    &, RHO_Cx , RHO_Cy, RHO_CZ , U , S , II_ITER , Tol , PHI )
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
   !Routine to solve Poisson equation on on uniform cell-center FV.
   !
   !
   !
   !
   !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
     INTEGER,INTENT(IN) :: Nx , Ny , NZ
 
     REAL(KIND=DP),INTENT(IN ) :: Dx , Dy , DZ , Dt , Uc(:,:,:) , Vc(:,:,:)  , Wc(:,:,:) , Tol
     REAL(KIND=DP),INTENT(IN ) :: S(:,:,:) , RHO_Cx(:,:,:) , RHO_Cy(:,:,:) , RHO_CZ(:,:,:) , PHI(:,:,:)
 
     INTEGER,INTENT(IN) :: II_ITER
 
     REAL(KIND=DP),INTENT(INOUT) :: U(:,:,:)
 
     REAL(KIND=DP),DIMENSION(Nx,Ny,NZ) :: U_old , U_old2 , SS
 
    
 
     REAL(KIND=DP) :: Res , W
     INTEGER :: I , J , K , Iter , N , Ierr
 
 
     W = 1.0_DP
     IF(II_ITER == 1) THEN
       W  = 1.9_DP
     END IF
 
     SS = 0.0_DP
 
 !$OMP PARALLEL DEFAULT(NONE) &
 !$OMP SHARED(SS,UC,VC,WC,Nx,NY,NZ) PRIVATE(I,J,K) FIRSTPRIVATE(DT,DX,DY,DZ) 
 !$OMP DO
 
     DO K = 2 , NZ-1
     DO J = 2 , NY-1
     DO I = 2 , NX-1
       SS(I,J,K) =   ( (Uc(I,J,K) - Uc(I-1,J,K)) / Dx + (Vc(I,J,K) - Vc(I,J-1,K)) / Dy   &
                      &  + (Wc(I,J,K) - Wc(I,J,K-1)) / DZ ) / Dt  
     END DO
     END DO
     END DO
     
 !$OMP END DO
 !$OMP END PARALLEL
 
      
   
  
      U_old = U
     
     
       DO K = 1 , NZ
       DO J = 1 , Ny
         U_old(1,J,K)  =  2.0_DP * U(1,J,K)  - U(2,J,K)   
         U_old(Nx,J,K) =  2.0_DP * U(Nx,J,K) -  U(Nx-1,J,K) 
    
        
       !   U_old(1,J)  = 2.0_DP * U(1,J)  - U(2,J)
       !   U_old(Nx,J) = 2.0_DP * U(Nx,J) - U(Nx-1,J)
    
       END DO
       END DO
         
       DO K = 1  ,NZ
       DO I = 1 , Nx
         U_old(I,1,K)  =  2.0_DP * U(I,1,K)  -  U(I,2,K)  
         U_old(I,Ny,K) =  2.0_DP * U(I,Ny,K) -  U(I,Ny-1,K) 
    
      !   U_old(I,1)  = 2.0_DP * U(I,1)  - U(I,2)
      !   U_old(I,Ny) = 2.0_DP * U(I,Ny) - U(I,Ny-1)
       END DO
       END DO
       
          
       DO J = 1 , NY
       DO I = 1 , Nx
         U_old(I,J,1)  =  2.0_DP * U(I,J,1)  -  U(I,J,2)  
         U_old(I,J,NZ) =  2.0_DP * U(I,J,NZ) -  U(I,J,NZ-1)      
       END DO
       END DO          
       
     
   
    
 
     U_old2 = U_old
 
 
 
     Iter = 0
     DO
       Iter = Iter + 1
 
  
 
 
 
 
  !$OMP PARALLEL DEFAULT(NONE) &
  !$OMP SHARED(U_OLD,SS,RHO_CX,RHO_CY,RHO_CZ,Nx,NY,NZ) PRIVATE(I,J,K) FIRSTPRIVATE(W,DX,DY,DZ) 
  !$OMP DO
         DO K = 2 , NZ-1
         DO J = 2 , Ny-1
         DO I = 2 , Nx-1
          U_old(I,J,K) = (1.0_DP - W) * U_old(I,J,K) &
           &          + W * ( -SS(I,J,K) &
                  &      +(RHO_Cx(I,J,K) * U_old(I+1,J,K) + RHO_Cx(I-1,J,K) * U_old(I-1,J,K)) / (Dx**2) &
                  &      +(RHO_Cy(I,J,K) * U_old(I,J+1,K) + RHO_Cy(I,J-1,K) * U_old(I,J-1,K)) / (Dy**2) &
                  &      +(RHO_Cz(I,J,K) * U_old(I,J,K+1) + RHO_Cz(I,J,K-1) * U_old(I,J,K-1)) / (DZ**2) &
                  &         ) / (  (RHO_Cx(I,J,K) + RHO_Cx(I-1,J,K)) / (Dx**2) + (RHO_CY(I,J,K) + RHO_CY(I,J-1,K)) / (DY**2) &
                  &              + (RHO_CZ(I,J,K) + RHO_CZ(I,J,K-1)) / (DZ**2) ) 
         END DO
         END DO
         END DO

  !$OMP END DO
  !$OMP END PARALLEL 
       
 
 
 
 !$OMP PARALLEL DEFAULT(NONE) &
 !$OMP SHARED(U_OLD,SS,RHO_CX,RHO_CY,RHO_CZ,Nx,NY,NZ) PRIVATE(I,J,K) FIRSTPRIVATE(W,DX,DY,DZ) 
 !$OMP DO      
  
          DO K = NZ-1 , 2 , -1
          DO J = Ny-1 , 2 , -1
          DO I = Nx-1 , 2 , -1
          	
             U_old(I,J,K) = (1.0_DP - W) * U_old(I,J,K) &
            &          + W * ( -SS(I,J,K) &
                   &      +(RHO_Cx(I,J,K) * U_old(I+1,J,K) + RHO_Cx(I-1,J,K) * U_old(I-1,J,K)) / (Dx**2) &
                   &      +(RHO_Cy(I,J,K) * U_old(I,J+1,K) + RHO_Cy(I,J-1,K) * U_old(I,J-1,K)) / (Dy**2) &
                   &      +(RHO_CZ(I,J,K) * U_old(I,J,K+1) + RHO_CZ(I,J,K-1) * U_old(I,J,K-1)) / (DZ**2) &
                   &         ) / (  (RHO_Cx(I,J,K) + RHO_Cx(I-1,J,K)) / (Dx**2) + (RHO_CY(I,J,K) + RHO_CY(I,J-1,K)) / (DY**2)  &
                   &              + (RHO_CZ(I,J,K) + RHO_CZ(I,J,K-1)) / (DZ**2) )
          END DO
          END DO
          END DO
 
 !$OMP END DO
 !$OMP END PARALLEL   
 
 
 
  
       DO K = 1 , NZ
       DO J = 1 , Ny
         U_old(1,J,K)  =   U_OLD(2,J,K)    
         U_old(Nx,J,K) =   U_OLD(Nx-1,J,K) 
       END DO
       END DO
         
       DO K = 1  ,NZ
       DO I = 1 , Nx
         U_old(I,1,K)  =   U_OLD(I,2,K) 
         U_old(I,Ny,K) =   U_OLD(I,Ny-1,K) 
       END DO
       END DO
              
       DO J = 1 , NY
       DO I = 1 , Nx
         U_old(I,J,1)  =   U_OLD(I,J,2)  
         U_old(I,J,NZ) =   U_OLD(I,J,NZ-1)         
       END DO
       END DO 
     
     
       U_old = U_old - U_old(Nx/2+1,Ny/2+1,NZ/2+1)
     
       Res = MAXVAL( ABS(U_old - U_old2) )
     
      IF(MOD(Iter , 100) == 0) THEN
        WRITE(*,*) ' G-S Iter , Res ' , Iter , Res
      END IF
      
     ! IF(MOD(II_Iter , 1) == 0) THEN
     !   WRITE(*,*) ' N-S  Iter , Res ' , II_Iter , Res
     ! END IF

     !!  N-S ITERATION  ONLY  II_ITER == 1  SOLVE PRESSURE POISSON
     ! IF(Iter == 10 .AND. II_ITER > 5) THEN
     !   EXIT
     ! END IF
      
      
      IF(Iter == 200 ) THEN
      !  WRITE(*,*) ' G-S Iter , Res ' , Iter , Res
        EXIT
      END IF

      IF(Res <= Tol*0.1_dp) THEN
      	
       !  WRITE(*,*) "convergence"
        EXIT
      END IF
 
       U_old2 = U_old
 
     END DO
 
 
     U = U_old
    
   END SUBROUTINE cell_face_POISSON_SOLVER_GS
   
   
    subroutine bc3d_bubble(fi,m,n,l)
	  
	  INTEGER,PARAMETER  :: DP = SELECTED_REAL_KIND(P=15)

	  INTEGER,INTENT(IN) :: m , n , L 
	  real(KIND=DP)  ::  fi(-2:m+3,-2:n+3,-2:l+3)
!
!     EXTRAPOLATED TO Ghost cells
!
!    i - dir
     DO J=1,N
     DO K=1,L
!        fi(0,J,K)=2.0*fi(1,J,K)-fi(2,J,K)
!        fi(-1,J,K)=2.0*fi(0,J,K)-fi(1,J,K)
!        fi(-2,J,K)=2.0*fi(-1,J,K)-fi(0,J,K)
!        fi(M+1,J,K)=2.0*fi(M,J,K)-fi(M-1,J,K)
!        fi(M+2,J,K)=2.0*fi(M+1,J,K)-fi(M,J,K)
!        fi(M+3,J,K)=2.0*fi(M+2,J,K)-fi(M+1,J,K)    

        fi(1,J,K)  = fi(2,J,K)   
        FI(1,J,K)=2.0_DP*Fi(1,J,K) - fi(2,j,K)
        fi(0,J,K)=fi(1,J,K)
        fi(-1,J,K)=fi(1,J,K)
        fi(-2,J,K)=fi(1,J,K)
        
      !  fi(m,J,K)  = ( 9.0_DP*fi(m-1,J,K)    - fi(m-2,J,K) ) / 8.0_DP
        fi(m,J,K)  = fi(m-1,J,K)   
        FI(M,J,K)=2.0_DP*Fi(M,J,K) - fi(m-1,j,K)
        fi(M+1,J,K)=fi(M,J,K)
        fi(M+2,J,K)=fi(M,J,K)
        fi(M+3,J,K)=fi(M,J,K)
      ENDDO
      ENDDO
      
      
!    j - dir
     DO I=-2,M+3
     DO K=1,L
!        fi(I,0,K)=2.0*fi(I,1,K)-fi(I,2,K)
!        fi(I,-1,K)=2.0*fi(I,0,K)-fi(I,1,K)
!        fi(I,-2,K)=2.0*fi(I,-1,K)-fi(I,0,K)
!        fi(I,N+1,K)=2.0*fi(I,N,K)-fi(I,N-1,K)
!        fi(I,N+2,K)=2.0*fi(I,N+1,K)-fi(I,N,K)
!        fi(I,N+3,K)=2.0*fi(I,N+2,K)-fi(I,N+1,K)  

     !   fi(I,1,K)  = ( 9.0_DP*fi(I,2,K)    - fi(I,3,K) ) / 8.0_DP  
     
        fi(I,1,K)  = fi(I,2,K)    
        FI(I,1,K)=2.0_DP*Fi(I,1,K) - fi(I,2,K) 
        fi(I,0,K)=fi(I,1,K)
        fi(I,-1,K)=fi(I,1,K)
        fi(I,-2,K)=fi(I,1,K)
        
      !  fi(I,N,K)  = ( 9.0_DP*fi(I,N-1,K)    - fi(I,N-2,K) ) / 8.0_DP
        fi(I,N,K)  = fi(I,N-1,K)   
        FI(I,N,K)=2.0_DP*Fi(I,N,K) - fi(I,N-1,K)
        fi(I,N+1,K)=fi(I,N,K)
        fi(I,N+2,K)=fi(I,N,K)
        fi(I,N+3,K)=fi(I,N,K)     
      ENDDO
      ENDDO
!    K - dir
     DO I=-2,M+3
     DO J=-2,N+3
!        fi(I,J,0)=2.0*fi(I,J,1)-fi(I,J,2)
!        fi(I,J,-1)=2.0*fi(I,J,0)-fi(I,J,1)
!        fi(I,J,-2)=2.0*fi(I,J,-1)-fi(I,J,0)
!        fi(I,J,L+1)=2.0*fi(I,J,L)-fi(I,J,L-1)
!        fi(I,J,L+2)=2.0*fi(I,J,L+1)-fi(I,J,L)
!        fi(I,J,L+3)=2.0*fi(I,J,L+2)-fi(I,J,L+1)   

      !  fi(I,J,1)  = fi(I,J,2)     
        
        
     !  IF( ( fi(I-1,j,2)*fi(I+1,j,2)) < 0.0_DP) THEN
           fi(I,j,1)  = fi(I,j,2) + COS(150.0_DP/180.0_DP*PI)*Dz / 2.0_DP   
     !  ELSE
     !    	 fi(I,j,1)  = fi(I,j,2)
     !  ENDIF 	 
        
        
        
        
        FI(I,J,1)=2.0_DP*Fi(I,J,1) - fi(I,j,2)  
        fi(I,J,0)=fi(I,J,1)
        fi(I,J,-1)=fi(I,J,1)
        fi(I,J,-2)=fi(I,J,1)
        
     !   fi(I,J,L)  = ( 9.0_DP*fi(I,J,L-1)    - fi(I,J,L-2) ) / 8.0_DP
        fi(I,J,L)  = fi(I,J,L-1)  
        FI(I,J,L)=2.0_DP*Fi(I,J,L) - fi(I,j,L-1)
        fi(I,J,L+1)=fi(I,J,L)
        fi(I,J,L+2)=fi(I,J,L)
        fi(I,J,L+3)=fi(I,J,L)
      ENDDO
      ENDDO
	 
	  end subroutine bc3d_bubble
	  
	  
	  
	    subroutine bc3d(fi,m,n,l)
	  
	  INTEGER,PARAMETER  :: DP = SELECTED_REAL_KIND(P=15)

	  INTEGER,INTENT(IN) :: m , n , L 
	  real(KIND=DP)  ::  fi(-2:m+3,-2:n+3,-2:l+3)
!
!     EXTRAPOLATED TO Ghost cells
!
!    i - dir
     DO J=1,N
     DO K=1,L
!        fi(0,J,K)=2.0*fi(1,J,K)-fi(2,J,K)
!        fi(-1,J,K)=2.0*fi(0,J,K)-fi(1,J,K)
!        fi(-2,J,K)=2.0*fi(-1,J,K)-fi(0,J,K)
!        fi(M+1,J,K)=2.0*fi(M,J,K)-fi(M-1,J,K)
!        fi(M+2,J,K)=2.0*fi(M+1,J,K)-fi(M,J,K)
!        fi(M+3,J,K)=2.0*fi(M+2,J,K)-fi(M+1,J,K)    

        fi(1,J,K)  = ( 9.0_DP*fi(2,J,K)    - fi(3,J,K) ) / 8.0_DP  
        FI(1,J,K)=2.0_DP*Fi(1,J,K) - fi(2,j,K)
        fi(0,J,K)=fi(1,J,K)
        fi(-1,J,K)=fi(1,J,K)
        fi(-2,J,K)=fi(1,J,K)
        
        fi(m,J,K)  = ( 9.0_DP*fi(m-1,J,K)    - fi(m-2,J,K) ) / 8.0_DP
        FI(M,J,K)=2.0_DP*Fi(M,J,K) - fi(m-1,j,K)
        fi(M+1,J,K)=fi(M,J,K)
        fi(M+2,J,K)=fi(M,J,K)
        fi(M+3,J,K)=fi(M,J,K)
      ENDDO
      ENDDO
!    j - dir
     DO I=-2,M+3
     DO K=1,L
!        fi(I,0,K)=2.0*fi(I,1,K)-fi(I,2,K)
!        fi(I,-1,K)=2.0*fi(I,0,K)-fi(I,1,K)
!        fi(I,-2,K)=2.0*fi(I,-1,K)-fi(I,0,K)
!        fi(I,N+1,K)=2.0*fi(I,N,K)-fi(I,N-1,K)
!        fi(I,N+2,K)=2.0*fi(I,N+1,K)-fi(I,N,K)
!        fi(I,N+3,K)=2.0*fi(I,N+2,K)-fi(I,N+1,K)  

        fi(I,1,K)  = ( 9.0_DP*fi(I,2,K)    - fi(I,3,K) ) / 8.0_DP  
        FI(I,1,K)=2.0_DP*Fi(I,1,K) - fi(I,2,K) 
        fi(I,0,K)=fi(I,1,K)
        fi(I,-1,K)=fi(I,1,K)
        fi(I,-2,K)=fi(I,1,K)
        
        fi(I,N,K)  = ( 9.0_DP*fi(I,N-1,K)    - fi(I,N-2,K) ) / 8.0_DP
        FI(I,N,K)=2.0_DP*Fi(I,N,K) - fi(I,N-1,K)
        fi(I,N+1,K)=fi(I,N,K)
        fi(I,N+2,K)=fi(I,N,K)
        fi(I,N+3,K)=fi(I,N,K)     
      ENDDO
      ENDDO
!    K - dir
     DO I=-2,M+3
     DO J=-2,N+3
!        fi(I,J,0)=2.0*fi(I,J,1)-fi(I,J,2)
!        fi(I,J,-1)=2.0*fi(I,J,0)-fi(I,J,1)
!        fi(I,J,-2)=2.0*fi(I,J,-1)-fi(I,J,0)
!        fi(I,J,L+1)=2.0*fi(I,J,L)-fi(I,J,L-1)
!        fi(I,J,L+2)=2.0*fi(I,J,L+1)-fi(I,J,L)
!        fi(I,J,L+3)=2.0*fi(I,J,L+2)-fi(I,J,L+1)   

        fi(I,J,1)  = ( 9.0_DP*fi(I,J,2)    - fi(I,J,3) ) / 8.0_DP  
        FI(I,J,1)=2.0_DP*Fi(I,J,1) - fi(I,j,2)  
        fi(I,J,0)=fi(I,J,1)
        fi(I,J,-1)=fi(I,J,1)
        fi(I,J,-2)=fi(I,J,1)
        
        fi(I,J,L)  = ( 9.0_DP*fi(I,J,L-1)    - fi(I,J,L-2) ) / 8.0_DP
        FI(I,J,L)=2.0_DP*Fi(I,J,L) - fi(I,j,L-1)
        fi(I,J,L+1)=fi(I,J,L)
        fi(I,J,L+2)=fi(I,J,L)
        fi(I,J,L+3)=fi(I,J,L)
      ENDDO
      ENDDO
	 
	  end subroutine bc3d
   
   
   
   
