SUBROUTINE TWIN_DEC( A, B, C , AA, BB , CC , AAA , BBB , CCC)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Routine to solve the CCD matrix by twin-elimination method.
  !
  !Input is the CCD matrix.
  !Output is the twin-forward elimination matrix.
  !Note that the coefficients for elimination
  ! A(1,:) , B(1,:) , AA(1,:) and BB(1,:)
  !are stored in the same locations.
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    REAL(KIND=DP), INTENT(INOUT), DIMENSION(:,:) :: A,  B  ,  C 
    REAL(KIND=DP), INTENT(INOUT), DIMENSION(:,:) :: AA, BB ,  CC
    REAL(KIND=DP), INTENT(INOUT), DIMENSION(:,:) :: AAA, BBB , CCC
   !--------------
    INTEGER       ::  i, n
    REAL(KIND=DP) ::  den , SC1 , SC2 , SC3

    n = SIZE(A,DIM=2)

    A (1,1)  = 0.0_DP
    AA(1,1)  = 0.0_DP
    AAA(1,1) = 0.0_DP
    A (3,n)  = 0.0_DP
    AA(3,n)  = 0.0_DP
    AAA(3,n) = 0.0_DP

    B (1,1)  = 0.0_DP
    BB(1,1)  = 0.0_DP
    BBB(1,1) = 0.0_DP
    B (3,n)  = 0.0_DP
    BB(3,n)  = 0.0_DP
    BBB(3,n) = 0.0_DP
    
    C(1,1)   = 0.0_DP
    CC(1,1)  = 0.0_DP
    CCC(1,1) = 0.0_DP
    C (3,n)  = 0.0_DP
    CC(3,n)  = 0.0_DP
    CCC(3,n) = 0.0_DP

    DO i = 2, N
    	! Den = A(2,i-1)*BB(2,i-1) - AA(2,i-1)*B(2,i-1)
    Den = ( A(2,i-1)*BB(2,i-1)*CCC(2,I-1) +  B(2,i-1)*CC(2,i-1)*AAA(2,I-1) + C(2,i-1)*AA(2,i-1)*BBB(2,I-1) ) &
      & -(  AAA(2,i-1)*BB(2,i-1)*C(2,I-1)  +  BBB(2,i-1)*CC(2,i-1)*A(2,I-1) +  CCC(2,i-1)*AA(2,i-1)*B(2,I-1)    )
     
     !!  SC1 = -( AA(2,i-1)*B(1,i) - A(1,i)*BB(2,i-1) )/Den  !!! FOR X
       
    SC1 =( ( A(1,I)*BB(2,I-1)*CCC(2,I-1) + B(1,I)*CC(2,I-1)*AAA(2,I-1) + C(1,I)*AA(2,I-1)*BBB(2,I-1)   )  &
    & -(  AAA(2,i-1)*BB(2,i-1)*C(1,I)  +  BBB(2,i-1)*CC(2,i-1)*A(1,I) +  CCC(2,i-1)*AA(2,i-1)*B(1,I)  )  ) / DEN
      
      
    !! SC2 =  (  B(1,i)*A(2,i-1) - B(2,i-1)*A(1,i)  )/Den   !!! for y
      
    SC2 = (  ( A(2,I-1)*B(1,I)*CCC(2,I-1) +  B(2,I-1)*C(1,I)*AAA(2,I-1) + C(2,I-1)*A(1,I)*BBB(2,I-1) )  &
    &	-(  AAA(2,i-1)*B(1,i)*C(2,I-1) + BBB(2,i-1)*C(1,i)*A(2,I-1)  + CCC(2,i-1)*A(1,i)*B(2,I-1)   )  ) /DEN
    	
     !! SC3    !!! for z
    	
    SC3  = ( ( A(2,I-1)*BB(2,I-1)*C(1,I) +   B(2,I-1)*CC(2,I-1)*A(1,I) + C(2,I-1)*AA(2,I-1)*B(1,I)  ) &
    &	-(  A(1,I)*BB(2,I-1)*C(2,I-1) +   B(1,I)*CC(2,I-1)*A(2,I-1) +  C(1,I)*AA(2,I-1)*B(2,I-1)   )  ) /DEN
 
    
     ! A(1,i) = SC1
     ! A(2,i) = A(2,i) - ( SC1*A(3,i-1) + SC2*AA(3,i-1) )

     ! B(1,i) = SC2
     ! B(2,i) = B(2,i) - ( SC1*B(3,i-1) + SC2*BB(3,i-1) )
    
     
      A(1,i) = SC1
      A(2,i) = A(2,i) - ( SC1*A(3,i-1) + SC2*AA(3,i-1) + SC3*AAA(3,I-1) )
      B(1,i) = SC2
      B(2,i) = B(2,i) - ( SC1*B(3,i-1) + SC2*BB(3,i-1) + SC3*BBB(3,I-1) )
      C(1,i) = SC3
      C(2,i) = C(2,i) - ( SC1*C(3,i-1) + SC2*CC(3,i-1) + SC3*CCC(3,I-1) )
  !
  ! =====================================================================   
           
     
    SC1 =( ( AA(1,I)*BB(2,I-1)*CCC(2,I-1) + BB(1,I)*CC(2,I-1)*AAA(2,I-1) + CC(1,I)*AA(2,I-1)*BBB(2,I-1)   )  &
    & -(  AAA(2,i-1)*BB(2,i-1)*CC(1,I)  +  BBB(2,i-1)*CC(2,i-1)*AA(1,I) +  CCC(2,i-1)*AA(2,i-1)*BB(1,I)  )  ) / DEN
      
      
    SC2 = (  ( A(2,I-1)*BB(1,I)*CCC(2,I-1) +  B(2,I-1)*CC(1,I)*AAA(2,I-1) + C(2,I-1)*AA(1,I)*BBB(2,I-1) )  &
    &	-(  AAA(2,i-1)*BB(1,i)*C(2,I-1) + BBB(2,i-1)*CC(1,i)*A(2,I-1)  + CCC(2,i-1)*AA(1,i)*B(2,I-1)   )  ) /DEN
    	
   	
    SC3  = ( ( A(2,I-1)*BB(2,I-1)*CC(1,I) +   B(2,I-1)*CC(2,I-1)*AA(1,I) + C(2,I-1)*AA(2,I-1)*BB(1,I)  ) &
    &	-(  AA(1,I)*BB(2,I-1)*C(2,I-1) +   BB(1,I)*CC(2,I-1)*A(2,I-1) +  CC(1,I)*AA(2,I-1)*B(2,I-1)   )  ) /DEN

    !  AA(1,i) = SC1
    !  AA(2,i) = AA(2,i) - (SC1*A(3,i-1) + SC2*AA(3,i-1))
    !
    !  BB(1,i) = SC2
    !  BB(2,i) = BB(2,i) - (SC1*B(3,i-1) + SC2*BB(3,i-1))
    
      AA(1,i) = SC1
      AA(2,i) = AA(2,i) - (SC1*A(3,i-1) + SC2*AA(3,i-1) + SC3*AAA(3,I-1))
      BB(1,i) = SC2
      BB(2,i) = BB(2,i) - (SC1*B(3,i-1) + SC2*BB(3,i-1) + SC3*BBB(3,i-1))  
      CC(1,i) = SC3
      CC(2,i) = CC(2,i) - ( SC1*C(3,i-1) + SC2*CC(3,i-1) + SC3*CCC(3,I-1) )
      
      
   ! ======================================================================================   
           
     
    SC1 =( ( AAA(1,I)*BB(2,I-1)*CCC(2,I-1) + BBB(1,I)*CC(2,I-1)*AAA(2,I-1) + CCC(1,I)*AA(2,I-1)*BBB(2,I-1)   )  &
    & -(  AAA(2,i-1)*BB(2,i-1)*CCC(1,I)  +  BBB(2,i-1)*CC(2,i-1)*AAA(1,I) +  CCC(2,i-1)*AA(2,i-1)*BBB(1,I)  )  ) / DEN
      
      
    SC2 = (  ( A(2,I-1)*BBB(1,I)*CCC(2,I-1) +  B(2,I-1)*CCC(1,I)*AAA(2,I-1) + C(2,I-1)*AAA(1,I)*BBB(2,I-1) )  &
    &	-(  AAA(2,i-1)*BBB(1,i)*C(2,I-1) + BBB(2,i-1)*CCC(1,i)*A(2,I-1)  + CCC(2,i-1)*AAA(1,i)*B(2,I-1)   )  ) /DEN
    	
   	
    SC3  = ( ( A(2,I-1)*BB(2,I-1)*CCC(1,I) +   B(2,I-1)*CC(2,I-1)*AAA(1,I) + C(2,I-1)*AA(2,I-1)*BBB(1,I)  ) &
    &	-(  AAA(1,I)*BB(2,I-1)*C(2,I-1) +   BBB(1,I)*CC(2,I-1)*A(2,I-1) +  CCC(1,I)*AA(2,I-1)*B(2,I-1)   )  ) /DEN

    !  AA(1,i) = SC1
    !  AA(2,i) = AA(2,i) - (SC1*A(3,i-1) + SC2*AA(3,i-1))
    !
    !  BB(1,i) = SC2
    !  BB(2,i) = BB(2,i) - (SC1*B(3,i-1) + SC2*BB(3,i-1))
    
      AAA(1,i) = SC1
      AAA(2,i) = AAA(2,i) - (SC1*A(3,i-1) + SC2*AA(3,i-1) + SC3*AAA(3,I-1))
      BBB(1,i) = SC2
      BBB(2,i) = BBB(2,i) - (SC1*B(3,i-1) + SC2*BB(3,i-1) + SC3*BBB(3,i-1))  
      CCC(1,i) = SC3
      CCC(2,i) = CCC(2,i) - ( SC1*C(3,i-1) + SC2*CC(3,i-1) + SC3*CCC(3,I-1) ) 
      
      
      
    END DO

  END SUBROUTINE TWIN_DEC
  
  
   SUBROUTINE TWIN_BKS( A, B, C , AA , BB,  CC,  AAA,  BBB , CCC ,  S , SS ,SSS )
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Routine to solve the CCD matrix by twin-elimination method.
  !
  !Input is the twin-forward elimination matrix and CCD source terms.
  !Output is solution.
  !Note that twin-forward elimination matrix is never changed.
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    REAL(KIND=DP), INTENT(IN   ), DIMENSION(:,:) :: A,  B ,C 
    REAL(KIND=DP), INTENT(IN   ), DIMENSION(:,:) :: AA, BB , CC
    REAL(KIND=DP), INTENT(IN   ), DIMENSION(:,:) :: AAA, BBB , CCC
    REAL(KIND=DP), INTENT(INOUT), DIMENSION(:) :: S , SS , SSS
   !--------------
    INTEGER       :: i, n
    REAL(KIND=DP) :: Den, SolS, SolSS ,SOLSSS

    n = SIZE(A,DIM=2)
   
   !! CALCUALTION SOURCE TERMS    A(1,I) = SC1 ,  B(1,I) = SC2 , C(1,I) = SC3
   

    DO i = 2, N
      S(i)  = S(i)  -(  A(1,i)*S(i-1) +  B(1,i)*SS(i-1) + C(1,i)*SSS(i-1)  )
      SS(i) = SS(i) -(  AA(1,i)*S(i-1) + BB(1,i)*SS(i-1) + CC(1,i)*SSS(i-1) )
      SSS(i) = SSS(i) -( AAA(1,i)*S(i-1) + BBB(1,i)*SS(i-1) + CCC(1,i)*SSS(i-1) )
    END DO
    
     

   !!! BACK UP FOR  N POINT
     
    Den   = (  A(2,n)*BB(2,n)*CCC(2,N) + AA(2,n)*BBB(2,n)*C(2,N) + AAA(2,N)*B(2,N)*CC(2,N)  ) &
            & - ( C(2,n)*BB(2,n)*AAA(2,N) + CC(2,n)*BBB(2,n)*A(2,N) + CCC(2,N)*B(2,N)*AA(2,N)    )
    
    
    SolS  = (  (  S(n)*BB(2,n)*CCC(2,N) + SS(n)*BBB(2,n)*C(2,N) + SSS(N)*B(2,N)*CC(2,N)  ) &
            & - ( C(2,n)*BB(2,n)*SSS(N) + CC(2,n)*BBB(2,n)*S(N) + CCC(2,N)*B(2,N)*SS(N)    )  ) / DEN
            
    SOLSS   = ( (  A(2,n)*SS(n)*CCC(2,N) + AA(2,n)*SSS(n)*C(2,N) + AAA(2,N)*S(N)*CC(2,N)  ) &
            & - ( C(2,n)*SS(n)*AAA(2,N) + CC(2,n)*SSS(n)*A(2,N) + CCC(2,N)*S(N)*AA(2,N)    )  ) /DEN      
    
    SOLSSS  =(  (  A(2,n)*BB(2,n)*SSS(N) + AA(2,n)*BBB(2,n)*S(N) + AAA(2,N)*B(2,N)*SS(N)  ) &
            & - ( S(n)*BB(2,n)*AAA(2,N) + SS(n)*BBB(2,n)*A(2,N) + SSS(N)*B(2,N)*AA(2,N)    )  ) /DEN
    
    
    S(n)  = SolS
    SS(n) = SolSS
    SSS(n) = SolSSS
    
   !!!  A(2,n)*BB(2,n)  Y*
   ! Den   = A(2,n)*BB(2,n) - AA(2,n)*B(2,n)
   ! SolS  = -( B(2,n)*SS(n) - BB(2,n)*S(n) )/Den
   ! SolSS =  ( A(2,n)*SS(n) - AA(2,n)*S(n) )/Den
   ! S(n)  = SolS
   ! SS(n) = SolSS
   
   
    !! !!! BACK UP FOR  N-1 , 1  , -1  POINT
   
    DO i = N-1, 1,-1
                     
    	S(i)  = S(i)  - (  A(3,i)*SolS +  B(3,i)*SolSS +   C(3,i)*SolSSS  )
    	SS(i) = SS(i) - ( AA(3,i)*SolS + BB(3,i)*SolSS + CC(3,i)*SolSSS )
    	SSS(i) = SSS(i) - ( AAA(3,i)*SolS + BBB(3,i)*SolSS + CCC(3,i)*SolSSS )

    Den   = (  A(2,I)*BB(2,I)*CCC(2,I) + AA(2,I)*BBB(2,I)*C(2,I) + AAA(2,I)*B(2,I)*CC(2,I)  ) &
            & - ( C(2,I)*BB(2,I)*AAA(2,I) + CC(2,I)*BBB(2,I)*A(2,I) + CCC(2,I)*B(2,I)*AA(2,I)    )
            
    SolS  = (  (  S(I)*BB(2,I)*CCC(2,I) + SS(I)*BBB(2,I)*C(2,I) + SSS(I)*B(2,I)*CC(2,I)  ) &
            & - ( C(2,I)*BB(2,I)*SSS(I) + CC(2,I)*BBB(2,I)*S(I) + CCC(2,I)*B(2,I)*SS(I)    )  ) / DEN
            
    SOLSS   = ( (  A(2,I)*SS(I)*CCC(2,I) + AA(2,I)*SSS(I)*C(2,I) + AAA(2,I)*S(I)*CC(2,I)  ) &
            & - ( C(2,I)*SS(I)*AAA(2,I) + CC(2,I)*SSS(I)*A(2,I) + CCC(2,I)*S(I)*AA(2,I)    )  ) /DEN      
    
    SOLSSS  =(  (  A(2,I)*BB(2,I)*SSS(I) + AA(2,I)*BBB(2,I)*S(I) + AAA(2,I)*B(2,I)*SS(I)  ) &
            & - ( S(I)*BB(2,I)*AAA(2,I) + SS(I)*BBB(2,I)*A(2,I) + SSS(I)*B(2,I)*AA(2,I)    )  ) /DEN
            
      S(i)  = SolS
      SS(i) = SolSS
      SSS(i) = SolSSS

    END DO

   ! DO i = N-1, 1,-1
   !
   ! 	S(i)  = S(i)  - (  A(3,i)*SolS +  B(3,i)*SolSS )
   ! 	SS(i) = SS(i) - ( AA(3,i)*SolS + BB(3,i)*SolSS )
   !
   !   Den   = A(2,i)*BB(2,i) - AA(2,i)*B(2,i)
   !   SolS  = -( B(2,i)*SS(i) - BB(2,i)*S(i))/Den
   !   SolSS =  ( A(2,i)*SS(i) - AA(2,i)*S(i))/Den
   !   S(i)  = SolS
   !   SS(i) = SolSS
   !
   ! END DO



  END SUBROUTINE TWIN_BKS
  
  
   SUBROUTINE CCD_UD_TWIN(Nx , Ny , Nz  , Dx , Dy  , Dz , C_X , C_Y , C_Z , U , V , W &
                         &, U_X , U_XX , U_XXX , V_Y , V_YY , V_YYY , W_Z , W_ZZ , W_ZZZ &
                         &, AU_X , BU_X , CU_X , AAU_X , BBU_X , CCU_X , AAAU_X , BBBU_X , CCCU_X&
                         &, AD_X , BD_X , CD_X , AAD_X , BBD_X , CCD_X , AAAD_X , BBBD_X , CCCD_X &
                         &, AU_Y , BU_Y , CU_Y , AAU_Y , BBU_Y , CCU_Y , AAAU_Y , BBBU_Y , CCCU_Y &
                         &, AD_Y , BD_Y , CD_Y , AAD_Y , BBD_Y , CCD_Y , AAAD_Y , BBBD_Y , CCCD_Y &
                         &, AU_Z , BU_Z , CU_Z , AAU_Z , BBU_Z , CCU_Z , AAAU_Z , BBBU_Z , CCCU_Z &
                         &, AD_Z , BD_Z , CD_Z , AAD_Z , BBD_Z , CCD_Z , AAAD_Z , BBBD_Z , CCCD_Z )
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Routine to solve U_X , U_XX , V_Y , V_YY by DRP-CCD.
  !
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    IMPLICIT NONE

    INTEGER,INTENT(IN) :: Nx , Ny , Nz
    REAL(KIND=DP),INTENT(IN) :: Dx , Dy , Dz
    REAL(KIND=DP),DIMENSION(:,:,:),INTENT(IN) :: C_X , C_Y , C_Z , U , V , W
    REAL(KIND=DP),DIMENSION(:,:,:),INTENT(OUT) :: U_X , U_XX  , U_XXX , V_Y , V_YY , V_YYY , W_Z , W_ZZ , W_ZZZ

    REAL(KIND=DP),DIMENSION(:,:),INTENT(IN) :: AU_X , BU_X , CU_X , AAU_X , BBU_X , CCU_X , AAAU_X , BBBU_X , CCCU_X 
    REAL(KIND=DP),DIMENSION(:,:),INTENT(IN) :: AD_X , BD_X , CD_X , AAD_X , BBD_X , CCD_X , AAAD_X , BBBD_X , CCCD_X

    REAL(KIND=DP),DIMENSION(:,:),INTENT(IN) :: AU_Y , BU_Y , CU_Y , AAU_Y , BBU_Y , CCU_Y , AAAU_Y , BBBU_Y , CCCU_Y
    REAL(KIND=DP),DIMENSION(:,:),INTENT(IN) :: AD_Y , BD_Y , CD_Y , AAD_Y , BBD_Y , CCD_Y , AAAD_Y , BBBD_Y , CCCD_Y
    
    REAL(KIND=DP),DIMENSION(:,:),INTENT(IN) :: AU_Z , BU_Z , CU_Z , AAU_Z , BBU_Z , CCU_Z , AAAU_Z , BBBU_Z , CCCU_Z
    REAL(KIND=DP),DIMENSION(:,:),INTENT(IN) :: AD_Z , BD_Z , CD_Z , AAD_Z , BBD_Z , CCD_Z , AAAD_Z , BBBD_Z , CCCD_Z

    REAL(KIND=DP) :: S(Nx,Ny,NZ) , SS(Nx,Ny,NZ) , SSS(Nx,Ny,NZ)
    REAL(KIND=DP) :: SD(Nx,Ny,NZ) , SSD(Nx,Ny,NZ) , SSSD(Nx,Ny,NZ)

    INTEGER :: I , J , K , Ierr


  
  !  DO J = 1 , Ny
  !  
  !    CALL ASSIGN_CCD_UD_TWIN_SRC(Nx , Dx , U(:,J) &
  !                             &, S(:,J) , SS(:,J) , SSS(:,J) , SD(:,J) , SSD(:,J) , SSSD(:,J))
  !
  !    !solve CCD matrix
  !    CALL TWIN_BKS( AU_X, BU_X, CU_X, AAU_X, BBU_X, CCU_X , AAAU_X, BBBU_X, CCCU_X, S(:,J) , SS(:,J)  , SSS(:,J) )
  !
  !    CALL TWIN_BKS( AD_X, BD_X , CD_X, AAD_X, BBD_X , CCD_X, AAAD_X, BBBD_X , CCCD_X, SD(:,J) , SSD(:,J) , SSSD(:,J) )
  !  END DO

    DO K = 1 , NZ
    DO J = 1 , Ny
    
      CALL ASSIGN_CCD_UD_TWIN_SRC(Nx , Dx , U(:,J,K) &
                               &, S(:,J,K) , SS(:,J,K) , SSS(:,J,K) , SD(:,J,K) , SSD(:,J,K) , SSSD(:,J,K))

      !solve CCD matrix
      CALL TWIN_BKS( AU_X, BU_X, CU_X, AAU_X, BBU_X, CCU_X , AAAU_X, BBBU_X, CCCU_X, S(:,J,K) , SS(:,J,K)  , SSS(:,J,K) )

      CALL TWIN_BKS( AD_X, BD_X , CD_X, AAD_X, BBD_X , CCD_X, AAAD_X, BBBD_X , CCCD_X, SD(:,J,K) , SSD(:,J,K) , SSSD(:,J,K) )
    END DO
    END DO



    DO K = 1 , NZ
    DO J = 1 , Ny
    DO I = 1 , Nx
      IF(C_X(I,J,K) >= 0.0_DP) THEN
        U_X(I,J,K)  = S(I,J,K)
      ELSE
        U_X(I,J,K)  = SD(I,J,K)
      END IF
       U_XX(I,J,K) = 0.5_DP*(SS(I,J,K) + SSD(I,J,K))
       U_XXX(I,J,K) = 0.5_DP*(SSS(I,J,K) + SSSD(I,J,K))
    END DO
    END DO
    END DO




  
   ! DO I = 1 , Nx
   ! 
   !   CALL ASSIGN_CCD_UD_TWIN_SRC(Ny , Dy , V(I,:) &
   !                            &, S(I,:) , SS(I,:) , SSS(I,:) , SD(I,:) , SSD(I,:) , SSSD(I,:))
   !
   !   !solve CCD matrix
   !   CALL TWIN_BKS( AU_Y, BU_Y, CU_Y, AAU_Y, BBU_Y , CCU_Y, AAAU_Y, BBBU_Y , CCCU_Y, S(I,:) , SS(I,:)  , SSS(I,:))
   !
   !   CALL TWIN_BKS( AD_Y, BD_Y, CD_Y, AAD_Y, BBD_Y, CCD_Y, AAAD_Y, BBBD_Y, CCCD_Y, SD(I,:) , SSD(I,:)  , SSSD(I,:) )
   ! END DO  
   
    DO K = 1 , NZ
    DO I = 1 , Nx
    
      CALL ASSIGN_CCD_UD_TWIN_SRC(Ny , Dy , V(I,:,K) &
                               &, S(I,:,K) , SS(I,:,K) , SSS(I,:,K) , SD(I,:,K) , SSD(I,:,K) , SSSD(I,:,K))

      !solve CCD matrix
      CALL TWIN_BKS( AU_Y, BU_Y, CU_Y, AAU_Y, BBU_Y , CCU_Y, AAAU_Y, BBBU_Y , CCCU_Y, S(I,:,K) , SS(I,:,K)  , SSS(I,:,K) )

      CALL TWIN_BKS( AD_Y, BD_Y, CD_Y, AAD_Y, BBD_Y, CCD_Y, AAAD_Y, BBBD_Y, CCCD_Y, SD(I,:,K) , SSD(I,:,K)  , SSSD(I,:,K) )
    END DO
    END DO



    DO K = 1 , NZ
    DO J = 1 , Ny
    DO I = 1 , Nx
        IF(C_Y(I,J,K) >= 0.0_DP) THEN
          V_Y(I,J,K)  = S(I,J,K)
        ELSE
          V_Y(I,J,K)  = SD(I,J,K)
        END IF
        V_YY(I,J,K) = 0.5_DP*(SS(I,J,K) + SSD(I,J,K))
        V_YYY(I,J,K) = 0.5_DP*(SSS(I,J,K) + SSSD(I,J,K))
    END DO
    END DO
    END DO
    
    
    
    DO J = 1 , NY
    DO I = 1 , Nx
    
      CALL ASSIGN_CCD_UD_TWIN_SRC(NZ , DZ , W(I,J,:) &
                               &, S(I,J,:) , SS(I,J,:) , SSS(I,J,:) , SD(I,J,:) , SSD(I,J,:) , SSSD(I,J,:))

      !solve CCD matrix
      CALL TWIN_BKS( AU_Z, BU_Z, CU_Z, AAU_Z, BBU_Z , CCU_Z, AAAU_Z, BBBU_Z , CCCU_Z, S(I,J,:) , SS(I,J,:)  , SSS(I,J,:))

      CALL TWIN_BKS( AD_Z, BD_Z, CD_Z, AAD_Z, BBD_Z, CCD_Z, AAAD_Z, BBBD_Z, CCCD_Z, SD(I,J,:) , SSD(I,J,:)  , SSSD(I,J,:) )
    END DO
    END DO



    DO K = 1 , NZ
    DO J = 1 , Ny
    DO I = 1 , Nx
        IF(C_Z(I,J,K) >= 0.0_DP) THEN
          W_Z(I,J,K)  = S(I,J,K)
        ELSE
          W_Z(I,J,K)  = SD(I,J,K)
        END IF
        W_ZZ(I,J,K) = 0.5_DP*(SS(I,J,K) + SSD(I,J,K))
        W_ZZZ(I,J,K) = 0.5_DP*(SSS(I,J,K) + SSSD(I,J,K))
    END DO
    END DO
    END DO


  END SUBROUTINE CCD_UD_TWIN
  
  
   SUBROUTINE ASSIGN_CCD_UD_TWIN(N , Dx , A , B , C ,  AA , BB  , CC , AAA , BBB , CCC &
                                      &, AD , BD , CD  ,  AAD , BBD , CCD  , AAAD , BBBD , CCCD )
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Routine to assign CCD coeff.
  !  AU , BU , CU are for the upwind coeff.
  !  AD , BD , CD are for the downwind coeff.
  !
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    IMPLICIT NONE

    INTEGER,INTENT(IN) :: N
    REAL(KIND=DP)                 ,INTENT(IN ) :: Dx
    REAL(KIND=DP),DIMENSION(:,:),INTENT(OUT) ::  A , B , C ,  AA , BB  , CC , AAA , BBB , CCC
    REAL(KIND=DP),DIMENSION(:,:),INTENT(OUT) ::  AD , BD , CD  ,  AAD , BBD , CCD  , AAAD , BBBD , CCCD

   ! REAL(KIND=DP) :: A1 , B1 , B2 , B3
   ! REAL(KIND=DP) :: A3D , B1D , B2D , B3D
   ! REAL(KIND=DP) :: AA1 , AA3 , BB1 , BB3
    
    REAL(KIND=DP) :: A1 , B1 , B2 , B3  , C1 ,C3
    REAL(KIND=DP) :: A3D , B1D , B2D , B3D , C1D  ,C3D
    REAL(KIND=DP) :: AA1 , AA3 , BB1 , BB3 , CC1 , CC3
    REAL(KIND=DP) :: AAA1 , AAA3 , BBB1 , BBB3 , CCC1 , CCC3

    INTEGER:: I
 
    
    
  !! FOR TWIN  7TH
    !For ux eq., a>0
    
   !    a1= 1.1875000000000000000_DP
   !    b1= 0.21830119013030133334_DP
   !    b2= -0.3774684542833426666_DP
   !    b3= -0.03169880986969866666_DP
   !    c1= 0.015540848149431500000_DP
   !    c3= 0.005292485183901833333_DP
   ! !!For ux eq., a<0    
   !    a3D= 1.1875000000000000000_DP
   !    b1D= 0.03169880986969866666_DP
   !    b2D= 0.3774684542833426666_DP
   !    b3D= -0.21830119013030133334_DP
   !    c1D= 0.005292485183901833333_DP         
   !    c3D= 0.015540848149431500000_DP 
       
       !!  KR >0         
          
         
         a1 = 1.1875000000000000000_DP 
         b1 = 0.23643236457259630771_DP      
         b2 = -0.2777469948507203076_DP 
         b3 = -0.01356763542740369229_DP    
         c1 = 0.018940443357361807695_DP 
         c3 = 0.001892889975971525638_DP 
        
         
       
      !! kR < 0   
      
         a3D = 1.1875000000000000000_DP  
         b1D = 0.01356763542740369229_DP
         b2D = 0.2777469948507203076_DP  
         b3D = -0.23643236457259630771_DP       
         c1D = 0.001892889975971525638_DP
         c3D = 0.018940443357361807695_DP     
      
                    
                                          

    !For uxx eq.
    AA1 = -29.0_DP / 16.0_DP
    AA3 =  29.0_DP / 16.0_DP
    BB1 = -5.0_DP / 16.0_DP
    BB3 = -5.0_DP / 16.0_DP
    CC1 = -1.0_DP / 48.0_DP
    CC3 =  1.0_DP / 48.0_DP
    
    
    !For uxxx eq.
    AAA1 =  -105.0_DP / 16.0_DP
    AAA3 =  -105.0_DP / 16.0_DP
    BBB1 =  -15.0_DP / 8.0_DP
    BBB3 =   15.0_DP / 8.0_DP
    CCC1 =  -3.0_DP / 16.0_DP
    CCC3 =  -3.0_DP / 16.0_DP
    
    

  !For Upwind
    !For  BOUNDARY
    !I POINT
     !!For U_X
     A(2,1) = 1.0_DP
     B(2,1) = 0.0_DP
     C(2,1) = 0.0_DP
     !!For U_XX
     AA(2,1) = 0.0_DP
     BB(2,1) = 1.0_DP
     CC(2,1) = 0.0_DP
     
    ! For U_XXX
     AAA(2,1) = 0.0_DP
     BBB(2,1) = 0.0_DP
     CCC(2,1) = 1.0_DP
     
     
    !I+1 POINT
     !For U_X
     A(3,1) =   4.0_DP      
     B(3,1) =   -1.0_DP*Dx
     C(3,1) =  (1.0_DP / 3.0_DP)*Dx**2
        
     !!!For U_XX
    ! AA(3,1) =  -20.0_DP / Dx
    ! BB(3,1) =  5.0_DP
    ! CC(3,1) =  (-7.0_DP / 3.0_DP)*DX
     
     AA(3,1) =  (-125.0_dp / 4.0_dp) /DX   
     BB(3,1) =    11.0_dp / 4.0_dp    
     CC(3,1) =  (-23.0_DP / 6.0_DP)*Dx   
     
     
    !!!For U_XXX
    ! AAA(3,1) =  60.0_DP / Dx**2
    ! BBB(3,1) =  -12.0_DP / DX
    ! CCC(3,1) =  9.0_DP
     
     AAA(3,1) =  (2485.0_DP / 12.0_DP) / Dx**2
     BBB(3,1) =  (95.0_DP/4.0_DP) / DX
     CCC(3,1) =  53.0_DP / 2.0_DP
     


    !For I = 2 , N-1 INNER CONDITION
    
    DO I = 2 , N-1
   !!!  I-1   !For U_X
    
      A(1,I) = A1
      B(1,I) = B1 * Dx
      C(1,I) = C1 * Dx**2
  ! !!!  I-1 For U_XX
      AA(1,I) = AA1 / Dx
      BB(1,I) = BB1                
      CC(1,I) = CC1 *DX
   !!!!  I-1  For U_XXX
      AAA(1,I) = AAA1 / Dx**2
      BBB(1,I) = BBB1 / DX
      CCC(1,I) = CCC1 
      
    !!!  I    !For U_X 
      
      A(2,I) = 1.0_DP
      B(2,I) = B2 * Dx
      C(2,I) = 0.0_DP * DX**2
    !!!  I    !For U_XX 
      AA(2,I) = 0.0_DP / DX
      BB(2,I) = 1.0_DP
      CC(2,I) = 0.0_DP * DX
    !!!  I    !For U_XXXX 
      AAA(2,I) = 0.0_DP   / Dx**2
      BBB(2,I) = 0.0_DP   / DX
      CCC(2,I) = 1.0_DP
      
     !!!!! !!!  I+1    !For U_X
     
      A(3,I) = 0.0_DP 
      B(3,I) = B3 *  Dx
      C(3,I) = C3 *DX**2
      !!!!! !!!  I+1    !For U_XX
      AA(3,I) = AA3 / Dx
      BB(3,I) = BB3 
      CC(3,I) = CC3 * DX
     !!!!! !!!  I+1    !For U_XXX
      AAA(3,I) = AAA3 / Dx**2
      BBB(3,I) = BBB3 / DX
      CCC(3,I) = CCC3
      
      
    END DO

    !For  !For  BOUNDARY I = N
    !I-1
     !For U_X
     A(1,N) = 4.0_DP
     B(1,N) = 1.0_DP *DX
     C(1,N) = ( 1.0_DP /3.0_DP )* Dx**2
     
     
     !For U_XX
    ! AA(1,N) = 20.0_DP / Dx
    ! BB(1,N) = 5.0_DP
    ! CC(1,N) = ( 7.0_DP / 3.0_DP ) *DX
     
     !For U_XX
     AA(1,N) = (125.0_DP/4.0_DP) / Dx
     BB(1,N) =  11.0_DP / 4.0_DP
     CC(1,N) = ( 23.0_DP / 6.0_DP ) *DX
     
    !For U_XXX
    ! AAA(1,N) = 60.0_DP / Dx**2
    ! BBB(1,N) = 12.0_DP  /DX
    ! CCC(1,N) = 9.0_DP
     
     !For U_XXX
     AAA(1,N) = (2485.0_DP/12.0_DP) / Dx**2
     BBB(1,N) = (-95.0_DP /4.0_DP)  /DX
     CCC(1,N) =  53.0_DP/2.0_DP
     
     
     
     
    !I
     !For U_X
     A(2,N) = 1.0_DP
     B(2,N) = 0.0_DP  *DX
     C(2,N) = 0.0_DP  *DX**2
     !For U_XX
     AA(2,N) = 0.0_DP  /DX
     BB(2,N) = 1.0_DP
     CC(2,N) = 0.0_DP  *DX
     
     !For U_XXX
     AAA(2,N) = 0.0_DP / DX**2
     BBB(2,N) = 0.0_DP / DX
     CCC(2,N) = 1.0_DP
     
     
     
     
     
 !================For Downwind For Downwind For Downwind For Downwind

  !For Downwind
    !For I = 1
      !!For U_X
     AD(2,1) = 1.0_DP
     BD(2,1) = 0.0_DP
     CD(2,1) = 0.0_DP
     !!For U_XX
     AAD(2,1) = 0.0_DP
     BBD(2,1) = 1.0_DP
     CCD(2,1) = 0.0_DP
     
    ! For U_XXX
     AAAD(2,1) = 0.0_DP
     BBBD(2,1) = 0.0_DP
     CCCD(2,1) = 1.0_DP
     
     
    !!!!!!!!!!!!!!I+1 POINT
     !For U_X
     AD(3,1) =  4.0_DP
     BD(3,1) = -1.0_DP*Dx
     CD(3,1) = (1.0_DP / 3.0_DP)*Dx**2
     
     
     !!!For U_XX
    ! AAD(3,1) =  -20.0_DP / Dx
    ! BBD(3,1) =  5.0_DP
    ! CCD(3,1) =  (-7.0_DP / 3.0_DP)*DX
     
     AAD(3,1) =  (-125.0_dp / 4.0_dp) /DX   
     BBD(3,1) =    11.0_dp / 4.0_dp    
     CCD(3,1) =  (-23.0_DP / 6.0_DP)*Dx   
     
     
    !!!For U_XXX
    ! AAAD(3,1) =  60.0_DP / Dx**2
    ! BBBD(3,1) =  -12.0_DP / DX
    ! CCCD(3,1) =  9.0_DP
    
     AAAD(3,1) =  (2485.0_DP / 12.0_DP) / Dx**2
     BBBD(3,1) =  (95.0_DP/4.0_DP) / DX
     CCCD(3,1) =  53.0_DP / 2.0_DP
     
     
     
    !For I = 2 , N-1 INNER CONDITION   !For I = 2 , N-1 INNER CONDITION  !For I = 2 , N-1 INNER CONDITION    
     
    DO I = 2 , N-1
   !!!  I-1   !For U_X
   
   
      AD(1,I) = 0.0_DP
      BD(1,I) = B1D * Dx
      CD(1,I) = C1D * Dx**2
  ! !!!  I-1 For U_XX
      AAD(1,I) = AA1 / Dx
      BBD(1,I) = BB1                
      CCD(1,I) = CC1 *DX
   !!!!  I-1  For U_XXX
      AAAD(1,I) = AAA1 / Dx**2
      BBBD(1,I) = BBB1 / DX
      CCCD(1,I) = CCC1 
      
   
      
      AD(2,I) = 1.0_DP
      BD(2,I) = B2D * Dx
      CD(2,I) = 0.0_DP * DX**2
    !!!  I    !For U_XX 
      AAD(2,I) = 0.0_DP / DX
      BBD(2,I) = 1.0_DP
      CCD(2,I) = 0.0_DP * DX
    !!!  I    !For U_XXXX 
      AAAD(2,I) = 0.0_DP   / Dx**2
      BBBD(2,I) = 0.0_DP   / DX
      CCCD(2,I) = 1.0_DP
      
     !!!!! !!!  I+1    !For U_X
     
     
     
      AD(3,I) = A3D 
      BD(3,I) = B3D *  Dx
      CD(3,I) = C3D *DX**2
      
      
      !!!!! !!!  I+1    !For U_XX
      AAD(3,I) = AA3 / Dx
      BBD(3,I) = BB3 
      CCD(3,I) = CC3 * DX
     !!!!! !!!  I+1    !For U_XXX
      AAAD(3,I) = AAA3 / Dx**2
      BBBD(3,I) = BBB3 / DX
      CCCD(3,I) = CCC3
      
      
    END DO
     
     
     
     
    

   !For  !For  BOUNDARY I = N
    !I-1
     !For U_X
     AD(1,N) = 4.0_DP
     BD(1,N) = 1.0_DP *DX
     CD(1,N) = ( 1.0_DP /3.0_DP )* Dx**2
     
     
     !For U_XX
    ! AAD(1,N) = 20.0_DP / Dx
    ! BBD(1,N) = 5.0_DP
    ! CCD(1,N) = ( 7.0_DP / 3.0_DP ) *DX
     
     AAD(1,N) = (125.0_DP/4.0_DP) / Dx
     BBD(1,N) =  11.0_DP / 4.0_DP
     CCD(1,N) = ( 23.0_DP / 6.0_DP ) *DX
     
    !For U_XXX
    ! AAAD(1,N) = 60.0_DP / Dx**2
    ! BBBD(1,N) = 12.0_DP  /DX
    ! CCCD(1,N) = 9.0_DP
     
     !For U_XXX
     AAAD(1,N) = (2485.0_DP/12.0_DP) / Dx**2
     BBBD(1,N) = (-95.0_DP /4.0_DP)  /DX
     CCCD(1,N) =  53.0_DP/2.0_DP
     
     
     
     
    !I
     !For U_X
     AD(2,N) = 1.0_DP
     BD(2,N) = 0.0_DP  *DX
     CD(2,N) = 0.0_DP  *DX**2
     !For U_XX
     AAD(2,N) = 0.0_DP  /DX
     BBD(2,N) = 1.0_DP
     CCD(2,N) = 0.0_DP  *DX
     
     !For U_XXX
     AAAD(2,N) = 0.0_DP / DX**2
     BBBD(2,N) = 0.0_DP / DX
     CCCD(2,N) = 1.0_DP

  END SUBROUTINE ASSIGN_CCD_UD_TWIN
  
  
  SUBROUTINE ASSIGN_CCD_UD_TWIN_SRC(N , Dx , U , S , SS , SSS ,  SD , SSD , SSSD)
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  !Routine to assign CCD source term.
  !  DU are for the upwind coeff.
  !  DD are for the downwind coeff.
  !
  !
  !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    IMPLICIT NONE

    INTEGER,INTENT(IN) :: N
    REAL(KIND=DP)             ,INTENT(IN ) :: Dx
    REAL(KIND=DP),DIMENSION(:),INTENT(IN ) :: U
    REAL(KIND=DP),DIMENSION(:),INTENT(OUT) :: S , SS  , SSS
    REAL(KIND=DP),DIMENSION(:),INTENT(OUT) :: SD , SSD ,  SSSD

    REAL(KIND=DP) :: D1 , D2 , D3
    REAL(KIND=DP) :: D1D , D2D , D3D
    REAL(KIND=DP) :: DD1 , DD2 , DD3
    REAL(KIND=DP) :: DDD1 , DDD2 , DDD3

    INTEGER:: I

      !   d1= -2.47211607402274_DP        
      !   d2= 2.7567321480454800000_DP    
      !   d3= -0.2846160740227400000_DP   
      !   
      !   d1D= 0.2846160740227400000_DP    
      !   d2D= -2.7567321480454800000_DP    
      !   d3D= 2.47211607402274_DP 
         
         
        d1 = -2.3361322657055276922_DP 
        d2 = 2.4847645314110553844_DP 
        d3 = -0.1486322657055276922_DP 
                      
        d1D = 0.1486322657055276922_DP
        d2D = -2.4847645314110553844_DP
        d3D = 2.3361322657055276922_DP    
       
        



    !For uxx eq.
    DD1 =  4.0_DP
    DD2 = -8.0_DP
    DD3 =  4.0_DP
    
    
    !For uxxx eq.
    DDD1 =  105.0_DP / 16.0_DP 
    DDD2 =  0.0_DP
    DDD3 =  -105.0_DP / 16.0_DP 
    

  !For upwind
    !For I = 1
     !U_X
      S(1)  = (-4.5_DP * U(1) + 4.0_DP * U(2) + 0.5_DP * U(3)) / Dx
     !U_XX
   !   SS(1) = ( 16.0_DP  * U(1) - 12.0_DP * U(2)  - 4.0_DP * U(3)  ) / (Dx**2)
   !  SSS(1) = ( -42.0_DP  * U(1) + 24.0_DP * U(2)  + 18.0_DP * U(3)  ) / (Dx**3)
     
     SS(1) = ( 19.0_DP  * U(1) + (-51.0_DP/8.0_DP) * U(2)  - 13.0_DP * U(3) + (3.0_DP/8.0_DP) * U(4)  ) / (Dx**2) 
    SSS(1) = ( (-289.0_DP/4.0_DP)  * U(1) + ( -5087.0_DP/72.0_DP ) * U(2)  + (303.0_DP/2.0_DP) * U(3)   &  
             &  + (-73.0_DP/8.0_DP) * U(4)  + ( 19.0_DP/36.0_DP) * U(5)      )    / (Dx**3)       
            

    !For I = 2 , N-1
    DO I = 2 , N-1
    	!U_X
      S(I)  = ( D1*U(I-1) + D2*U(I) + D3*U(I+1) ) / Dx
      !U_XX
      SS(I) = ( DD1*U(I-1) + DD2*U(I) + DD3*U(I+1) ) / (Dx**2)
      !U_XXX
      SSS(I) = ( DDD1*U(I-1) + DDD2*U(I) + DDD3*U(I+1) ) / (Dx**3)
    END DO
    
    

    !For I = N
     !U_X
      S(N)  = ( 4.5_DP * U(N) - 4.0_DP * U(N-1) - 0.5_DP * U(N-2) ) / Dx
     !U_XX
   !  SS(N) = ( 16.0_DP  * U(N) - 12.0_DP * U(N-1)  - 4.0_DP * U(N-2)  ) / (Dx**2)
   !  SSS(N) = ( 42.0_DP  * U(N) - 24.0_DP * U(N-1)  - 18.0_DP * U(N-2)  ) / (Dx**3)
     
    
    SS(N) = ( 19.0_DP  * U(N) + (-51.0_DP/8.0_DP)* U(N-1)  - 13.0_DP * U(N-2)  + (3.0_DP/8.0_DP) * U(N-3) ) / (Dx**2)
    SSS(N) = ( (289.0_DP/4.0_DP)  * U(N) +  (5087.0_DP / 72.0_DP ) * U(N-1)  +  (-303.0_DP/2.0_DP) * U(N-2) &
            &  +  (73.0_DP/8.0_DP) * U(N-3)  +  (-19.0_DP/36.0_DP) * U(N-4)    ) / (Dx**3)



  !For Downwind
    !For I = 1
     !U_X
      SD(1)  = (-4.5_DP * U(1) + 4.0_DP * U(2) + 0.5_DP * U(3)) / Dx
     !U_XX
     ! SSD(1) = ( 16.0_DP  * U(1) - 12.0_DP * U(2)  - 4.0_DP * U(3)  ) / (Dx**2)
     ! SSSD(1) = ( -42.0_DP  * U(1) + 24.0_DP * U(2)  + 18.0_DP * U(3)  ) / (Dx**3)
      
    SSD(1) = ( 19.0_DP  * U(1) + (-51.0_DP/8.0_DP) * U(2)  - 13.0_DP * U(3) + (3.0_DP/8.0_DP) * U(4)  ) / (Dx**2)
    SSSD(1) = ( (-289.0_DP/4.0_DP)  * U(1) + ( -5087.0_DP/72.0_DP ) * U(2)  + (303.0_DP/2.0_DP) * U(3)   &  
             &  + (-73.0_DP/8.0_DP) * U(4)  + ( 19.0_DP/36.0_DP) * U(5)      )    / (Dx**3)      
      
      

    !For I = 2 , N-1
    DO I = 2 , N-1
    	!U_X
      SD(I)  = ( D1D*U(I-1) + D2D*U(I) + D3D*U(I+1) ) / Dx
      !U_XX
      SSD(I) = ( DD1*U(I-1) + DD2*U(I) + DD3*U(I+1) ) / (Dx**2)
      
      !U_XXX
      SSSD(I) = ( DDD1*U(I-1) + DDD2*U(I) + DDD3*U(I+1) ) / (Dx**3)
    END DO

    !For I = N
     !U_X
      SD(N)  = ( 4.5_DP * U(N) - 4.0_DP * U(N-1) - 0.5_DP * U(N-2) ) / Dx
     
    !  SSD(N) = ( 16.0_DP  * U(N) - 12.0_DP * U(N-1)  - 4.0_DP * U(N-2)  ) / (Dx**2)
    !  SSSD(N) = ( 42.0_DP  * U(N) - 24.0_DP * U(N-1)  - 18.0_DP * U(N-2)  ) / (Dx**3)
     
     
     SSD(N) = ( 19.0_DP  * U(N) + (-51.0_DP/8.0_DP)* U(N-1)  - 13.0_DP * U(N-2)  + (3.0_DP/8.0_DP) * U(N-3) ) / (Dx**2)
     SSSD(N) = ( (289.0_DP/4.0_DP)  * U(N) +  (5087.0_DP / 72.0_DP ) * U(N-1)  +  (-303.0_DP/2.0_DP) * U(N-2) &
            &  +  (73.0_DP/8.0_DP) * U(N-3)  +  (-19.0_DP/36.0_DP) * U(N-4)    ) / (Dx**3)  

  END SUBROUTINE ASSIGN_CCD_UD_TWIN_SRC
  
  
 
  
  
  